"""Pytorch Geometric Dataset classes for Protein Graphs."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import os
from pathlib import Path
from typing import Callable, Dict, Generator, List, Optional

import networkx as nx
from loguru import logger as log
from tqdm import tqdm

from graphein.ml.conversion import GraphFormatConvertor
from graphein.protein.config import ProteinGraphConfig
from graphein.protein.graphs import construct_graphs_mp
from graphein.protein.utils import (
    download_alphafold_structure,
    download_pdb,
    download_pdb_multiprocessing,
)
from graphein.utils.dependencies import import_message

try:
    import torch
    from torch_geometric.data import Data, Dataset, InMemoryDataset
except ImportError:
    import_message(
        "graphein.ml.datasets.torch_geometric_dataset",
        "torch_geometric",
        conda_channel="pyg",
        pip_install=True,
    )


class InMemoryProteinGraphDataset(InMemoryDataset):
    def __init__(
        self,
        root: str,
        name: str,
        paths: Optional[List[str]] = None,
        pdb_codes: Optional[List[str]] = None,
        uniprot_ids: Optional[List[str]] = None,
        graph_label_map: Optional[Dict[str, torch.Tensor]] = None,
        node_label_map: Optional[Dict[str, torch.Tensor]] = None,
        chain_selection_map: Optional[Dict[str, List[str]]] = None,
        graphein_config: ProteinGraphConfig = ProteinGraphConfig(),
        graph_format_convertor: GraphFormatConvertor = GraphFormatConvertor(
            src_format="nx", dst_format="pyg"
        ),
        graph_transformation_funcs: Optional[List[Callable]] = None,
        transform: Optional[Callable] = None,
        pdb_transform: Optional[List[Callable]] = None,
        pre_transform: Optional[Callable] = None,
        pre_filter: Optional[Callable] = None,
        num_cores: int = 16,
        af_version: int = 2,
    ):
        """In Memory dataset for protein graphs.

        Dataset base class for creating graph datasets which easily fit
        into CPU memory. Inherits from
        :class:`torch_geometric.data.InMemoryDataset`, which inherits from
        :class:`torch_geometric.data.Dataset`.
        See `here <https://pytorch-geometric.readthedocs.io/en/latest/notes/
        create_dataset.html#creating-in-memory-datasets>`__ for the accompanying
        tutorial.

        :param root: Root directory where the dataset should be saved.
        :type root: str
        :param name: Name of the dataset. Will be saved to ``data_$name.pt``.
        :type name: str
        :param paths: List of full path of PDB or MMTF files to load. Defaults
            to ``None``.
        :type paths: Optional[List[str]], optional
        :param pdb_codes: List of PDB codes to download and parse from the PDB.
            Defaults to None.
        :type pdb_codes: Optional[List[str]], optional
        :param uniprot_ids: List of Uniprot IDs to download and parse from
            Alphafold Database. Defaults to ``None``.
        :type uniprot_ids: Optional[List[str]], optional
        :param graph_label_map: Dictionary mapping PDB/Uniprot IDs to
            graph-level labels. Defaults to ``None``.
        :type graph_label_map: Optional[Dict[str, Tensor]], optional
        :param node_label_map: Dictionary mapping PDB/Uniprot IDs to node-level
            labels. Defaults to ``None``.
        :type node_label_map: Optional[Dict[str, torch.Tensor]], optional
        :param chain_selection_map: Dictionary mapping, defaults to ``None``.
        :type chain_selection_map: Optional[Dict[str, List[str]]], optional
        :param graphein_config: Protein graph construction config, defaults to
            ``ProteinGraphConfig()``.
        :type graphein_config: ProteinGraphConfig, optional
        :param graph_format_convertor: Conversion handler for graphs, defaults
            to ``GraphFormatConvertor(src_format="nx", dst_format="pyg")``.
        :type graph_format_convertor: GraphFormatConvertor, optional
        :param pdb_transform: List of functions that consume a list of paths to
            the downloaded structures. This provides an entry point to apply
            pre-processing from bioinformatics tools of your choosing. Defaults
            to ``None``.
        :type pdb_transform: Optional[List[Callable]], optional
        :param graph_transformation_funcs: List of functions that consume a
            ``nx.Graph`` and return a ``nx.Graph``. Applied to graphs after
            construction but before conversion to pyg. Defaults to ``None``.
        :type graph_transformation_funcs: Optional[List[Callable]], optional
        :param transform: A function/transform that takes in a
            ``torch_geometric.data.Data`` object and returns a transformed
            version. The data object will be transformed before every access.
            Defaults to ``None``.
        :type transform: Optional[Callable], optional
        :param pre_transform:  A function/transform that takes in an
            ``torch_geometric.data.Data`` object and returns a transformed
            version. The data object will be transformed before being saved to
            disk. Defaults to ``None``.
        :type pre_transform: Optional[Callable], optional
        :param pre_filter:  A function that takes in a
            ``torch_geometric.data.Data`` object and returns a boolean value,
            indicating whether the data object should be included in the final
            dataset. Optional, defaults to ``None``.
        :type pre_filter: Optional[Callable], optional
        :param num_cores: Number of cores to use for multiprocessing of graph
            construction, defaults to ``16``.
        :type num_cores: int, optional
        :param af_version: Version of AlphaFoldDB structures to use,
            defaults to ``2``.
        :type af_version: int, optional
        """
        self.name = name
        self.pdb_codes = (
            [pdb.lower() for pdb in pdb_codes]
            if pdb_codes is not None
            else None
        )
        self.uniprot_ids = (
            [up.upper() for up in uniprot_ids]
            if uniprot_ids is not None
            else None
        )

        self.paths = paths
        if self.paths is None:
            if self.pdb_codes and self.uniprot_ids:
                self.structures = self.pdb_codes + self.uniprot_ids
            elif self.pdb_codes:
                self.structures = pdb_codes
            elif self.uniprot_ids:
                self.structures = uniprot_ids
        # Use local saved pdb_files instead of download or move them to
        # self.root/raw dir
        else:
            if isinstance(self.paths, list):
                self.structures = [
                    os.path.splitext(os.path.split(path)[-1])[0]
                    for path in self.paths
                ]
                self.path, _ = os.path.split(self.paths[0])

        if self.pdb_codes and self.uniprot_ids:
            self.structures = self.pdb_codes + self.uniprot_ids
        elif self.pdb_codes:
            self.structures = pdb_codes
        elif self.uniprot_ids:
            self.structures = uniprot_ids
        self.af_version = af_version
        self.bad_pdbs: List[str] = (
            []
        )  # list of pdb codes that failed to download

        # Labels & Chains
        self.graph_label_map = graph_label_map
        self.node_label_map = node_label_map
        self.chain_selection_map = chain_selection_map

        # Configs
        self.config = graphein_config
        self.graph_format_convertor = graph_format_convertor
        self.graph_transformation_funcs = graph_transformation_funcs
        self.pdb_transform = pdb_transform
        self.num_cores = num_cores
        self.af_version = af_version
        super().__init__(
            root,
            transform=transform,
            pre_transform=pre_transform,
            pre_filter=pre_filter,
        )
        self.config.pdb_dir = Path(self.raw_dir)
        self.data, self.slices = torch.load(
            self.processed_paths[0], weights_only=False
        )

    @property
    def raw_file_names(self) -> List[str]:
        """Name of the raw files in the dataset."""
        return [f"{pdb}.pdb" for pdb in self.structures]

    @property
    def processed_file_names(self) -> List[str]:
        """Name of the processed file."""
        return [f"data_{self.name}.pt"]

    @property
    def raw_dir(self) -> str:
        if self.paths is not None:
            return self.path  # replace raw dir with user local path
        else:
            return os.path.join(self.root, "raw")

    def download(self):
        """Download the PDB files from RCSB or Alphafold."""
        self.config.pdb_dir = Path(self.raw_dir)
        if self.pdb_codes:
            # Only download PDBs that are not already downloaded
            to_download = [
                pdb
                for pdb in set(self.pdb_codes)
                if not os.path.exists(Path(self.raw_dir) / f"{pdb}.pdb")
            ]
            download_pdb_multiprocessing(
                to_download,
                self.raw_dir,
                max_workers=self.num_cores,
                strict=False,
            )
            self.bad_pdbs = self.bad_pdbs + [
                pdb
                for pdb in set(self.pdb_codes)
                if not os.path.exists(Path(self.raw_dir) / f"{pdb}.pdb")
            ]
        if self.uniprot_ids:
            [
                download_alphafold_structure(
                    uniprot,
                    out_dir=self.raw_dir,
                    version=self.af_version,
                    aligned_score=False,
                )
                for uniprot in tqdm(self.uniprot_ids)
            ]

    def __len__(self) -> int:
        return len(self.structures)

    def transform_pdbs(self):
        """
        Performs pre-processing of PDB structures before constructing graphs.
        """
        structure_files = [
            f"{self.raw_dir}/{pdb}.pdb" for pdb in self.structures
        ]
        for func in self.pdb_transform:
            func(structure_files)

    def process(self):
        """Process structures into PyG format and save to disk."""
        # Read data into huge `Data` list.
        structure_files = [
            f"{self.raw_dir}/{pdb}.pdb" for pdb in self.structures
        ]

        # Apply transformations to raw PDB files.
        if self.pdb_transform is not None:
            print("Transforming PDB files...")
            self.transform_pdbs()

        if self.chain_selection_map:
            chain_selections = [
                (
                    self.chain_selection_map[pdb]
                    if pdb in self.chain_selection_map.keys()
                    else "all"
                )
                for pdb in self.structures
            ]
        else:
            chain_selections = None

        # Create graph objects
        print("Constructing Graphs...")
        graphs = construct_graphs_mp(
            path_it=structure_files,
            config=self.config,
            chain_selections=chain_selections,
            return_dict=True,
            num_cores=self.num_cores,
        )
        # Transform graphs
        if self.graph_transformation_funcs is not None:
            print("Transforming Nx Graphs...")
            for func in self.graph_transformation_funcs:
                graphs = {k: func(v) for k, v in graphs.items()}

        # Convert to PyTorch Geometric Data
        print("Converting Graphs...")
        graphs = {k: self.graph_format_convertor(v) for k, v in graphs.items()}
        graphs = dict(zip(self.structures, graphs.values()))

        # Assign labels
        if self.graph_label_map:
            print("Assigning graph Labels...")
            for k, v in self.graph_label_map.items():
                try:
                    graphs[k].graph_y = v
                except KeyError:
                    print(f"{k} not found in graphs. Skipping.")
        if self.node_label_map:
            print("Assigning node Labels...")
            for k, v in self.node_label_map.items():
                try:
                    graphs[k].node_y = v
                except KeyError:
                    print(f"{k} not found in graphs. Skipping.")
        data_list = list(graphs.values())
        del graphs

        if self.pre_filter is not None:
            print("Pre-filtering Data...")
            data_list = [g for g in data_list if self.pre_filter(g)]

        if self.pre_transform is not None:
            print("Pre-transforming data...")
            data_list = [self.pre_transform(data) for data in data_list]

        print("Saving Data...")
        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[0])
        print("Done!")


class ProteinGraphDataset(Dataset):
    def __init__(
        self,
        root: str,
        paths: Optional[List[str]] = None,
        pdb_codes: Optional[List[str]] = None,
        uniprot_ids: Optional[List[str]] = None,
        # graph_label_map: Optional[Dict[str, int]] = None,
        graph_labels: Optional[List[torch.Tensor]] = None,
        node_labels: Optional[List[torch.Tensor]] = None,
        chain_selections: Optional[List[str]] = None,
        # node_label_map: Optional[Dict[str, int]] = None,
        # chain_selection_map: Optional[Dict[str, List[str]]] = None,
        graphein_config: ProteinGraphConfig = ProteinGraphConfig(),
        graph_format_convertor: GraphFormatConvertor = GraphFormatConvertor(
            src_format="nx", dst_format="pyg"
        ),
        graph_transformation_funcs: Optional[List[Callable]] = None,
        pdb_transform: Optional[List[Callable]] = None,
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
        pre_filter: Optional[Callable] = None,
        num_cores: int = 16,
        af_version: int = 2,
    ):
        """Dataset class for protein graphs.

        Dataset base class for creating graph datasets.
        See `here <https://pytorch-geometric.readthedocs.io/en/latest/notes/
        create_dataset.html>`__ for the accompanying tutorial.

        :param root: Root directory where the dataset should be saved.
        :type root: str
        :param paths: List of full path of PDB or MMTF files to load. Defaults
            to ``None``.
        :type paths: Optional[List[str]], optional
        :param pdb_codes: List of PDB codes to download and parse from the PDB.
            Defaults to ``None``.
        :type pdb_codes: Optional[List[str]], optional
        :param uniprot_ids: List of Uniprot IDs to download and parse from
            Alphafold Database. Defaults to ``None``.
        :type uniprot_ids: Optional[List[str]], optional
        :param graph_label_map: Dictionary mapping PDB/Uniprot IDs to
            graph-level labels. Defaults to ``None``.
        :type graph_label_map: Optional[Dict[str, Tensor]], optional
        :param node_label_map: Dictionary mapping PDB/Uniprot IDs to node-level
            labels. Defaults to ``None``.
        :type node_label_map: Optional[Dict[str, torch.Tensor]], optional
        :param chain_selection_map: Dictionary mapping, defaults to ``None``.
        :type chain_selection_map: Optional[Dict[str, List[str]]], optional
        :param graphein_config: Protein graph construction config, defaults to
            ``ProteinGraphConfig()``.
        :type graphein_config: ProteinGraphConfig, optional
        :param graph_format_convertor: Conversion handler for graphs, defaults
            to ``GraphFormatConvertor(src_format="nx", dst_format="pyg")``.
        :type graph_format_convertor: GraphFormatConvertor, optional
        :param graph_transformation_funcs: List of functions that consume a
            ``nx.Graph`` and return a ``nx.Graph``. Applied to graphs after
            construction but before conversion to pyg. Defaults to ``None``.
        :type graph_transformation_funcs: Optional[List[Callable]], optional
        :param pdb_transform: List of functions that consume a list of paths to
            the downloaded structures. This provides an entry point to apply
            pre-processing from bioinformatics tools of your choosing. Defaults
            to ``None``.
        :type pdb_transform: Optional[List[Callable]], optional
        :param transform: A function/transform that takes in a
            ``torch_geometric.data.Data`` object and returns a transformed
            version. The data object will be transformed before every access.
            Defaults to ``None``.
        :type transform: Optional[Callable], optional
        :param pre_transform:  A function/transform that takes in an
            ``torch_geometric.data.Data`` object and returns a transformed
            version. The data object will be transformed before being saved to
            disk. Defaults to ``None``.
        :type pre_transform: Optional[Callable], optional
        :param pre_filter:  A function that takes in a
            ``torch_geometric.data.Data`` object and returns a boolean value,
            indicating whether the data object should be included in the final
            dataset. Optional, defaults to ``None``.
        :type pre_filter: Optional[Callable], optional
        :param num_cores: Number of cores to use for multiprocessing of graph
            construction, defaults to ``16``.
        :type num_cores: int, optional
        :param af_version: Version of AlphaFoldDB structures to use,
            defaults to ``2``.
        :type af_version: int, optional
        """
        self.pdb_codes = (
            [pdb.lower() for pdb in pdb_codes]
            if pdb_codes is not None
            else None
        )
        self.uniprot_ids = (
            [up.upper() for up in uniprot_ids]
            if uniprot_ids is not None
            else None
        )
        self.paths = paths
        if self.paths is None:
            if self.pdb_codes and self.uniprot_ids:
                self.structures = self.pdb_codes + self.uniprot_ids
            elif self.pdb_codes:
                self.structures = pdb_codes
            elif self.uniprot_ids:
                self.structures = uniprot_ids
        # Use local saved pdb_files instead of download or move them to
        # self.root/raw dir
        else:
            if isinstance(self.paths, list):
                self.structures = [
                    os.path.splitext(os.path.split(path)[-1])[0]
                    for path in self.paths
                ]
                self.path, _ = os.path.split(self.paths[0])

        # Labels & Chains

        self.examples: Dict[int, str] = dict(enumerate(self.structures))

        if graph_labels is not None:
            self.graph_label_map = dict(enumerate(graph_labels))
        else:
            self.graph_label_map = None

        if node_labels is not None:
            self.node_label_map = dict(enumerate(node_labels))
        else:
            self.node_label_map = None

        if chain_selections is not None:
            self.chain_selection_map = dict(enumerate(chain_selections))
        else:
            self.chain_selection_map = None
        self.validate_input()
        self.bad_pdbs: List[str] = []

        # Configs
        self.config = graphein_config
        self.graph_format_convertor = graph_format_convertor
        self.num_cores = num_cores
        self.pdb_transform = pdb_transform
        self.graph_transformation_funcs = graph_transformation_funcs
        self.af_version = af_version
        super().__init__(
            root,
            transform=transform,
            pre_transform=pre_transform,
            pre_filter=pre_filter,
        )
        self.config.pdb_dir = Path(self.raw_dir)

    @property
    def raw_file_names(self) -> List[str]:
        """Names of raw files in the dataset."""
        return [f"{pdb}.pdb" for pdb in self.structures]

    @property
    def processed_file_names(self) -> List[str]:
        """Names of processed files to look for"""
        if self.chain_selection_map is not None:
            return [
                f"{pdb}_{chain}.pt"
                for pdb, chain in zip(
                    self.structures, self.chain_selection_map.values()
                )
            ]
        else:
            return [f"{pdb}.pt" for pdb in self.structures]

    @property
    def raw_dir(self) -> str:
        if self.paths is not None:
            return self.path  # replace raw dir with user local path
        else:
            return os.path.join(self.root, "raw")

    def validate_input(self):
        if self.graph_label_map is not None:
            assert len(self.structures) == len(
                self.graph_label_map
            ), "Number of proteins and graph labels must match"
        if self.node_label_map is not None:
            assert len(self.structures) == len(
                self.node_label_map
            ), "Number of proteins and node labels must match"
        if self.chain_selection_map is not None:
            assert len(self.structures) == len(
                self.chain_selection_map
            ), "Number of proteins and chain selections must match"
            assert len(
                {
                    f"{pdb}_{chain}"
                    for pdb, chain in zip(
                        self.structures, self.chain_selection_map
                    )
                }
            ) == len(self.structures), "Duplicate protein/chain combinations"

    def download(self):
        """Download the PDB files from RCSB or Alphafold."""
        self.config.pdb_dir = Path(self.raw_dir)
        if self.pdb_codes:
            # Only download undownloaded PDBs
            to_download = [
                pdb
                for pdb in set(self.pdb_codes)
                if not os.path.exists(Path(self.raw_dir) / f"{pdb}.pdb")
            ]
            download_pdb_multiprocessing(
                to_download,
                self.raw_dir,
                max_workers=self.num_cores,
                strict=False,
            )
            self.bad_pdbs = self.bad_pdbs + [
                pdb
                for pdb in set(self.pdb_codes)
                if not os.path.exists(Path(self.raw_dir) / f"{pdb}.pdb")
            ]
        if self.uniprot_ids:
            to_download = [
                uniprot
                for uniprot in set(self.uniprot_ids)
                if not os.path.exists(Path(self.raw_dir) / f"{uniprot}.pdb")
            ]
            [
                download_alphafold_structure(
                    uniprot,
                    out_dir=self.raw_dir,
                    version=self.af_version,
                    aligned_score=False,
                )
                for uniprot in tqdm(to_download)
            ]

    def len(self) -> int:
        """Returns length of data set (number of structures)."""
        return len(self.structures)

    def transform_pdbs(self):
        """
        Performs pre-processing of PDB structures before constructing graphs.
        """
        structure_files = [
            f"{self.raw_dir}/{pdb}.pdb" for pdb in self.structures
        ]
        for func in self.pdb_transform:
            func(structure_files)

    def transform_graphein_graphs(self, graph: nx.Graph):
        for func in self.graph_transformation_funcs:
            graph = func(graph)
        return graph

    def process(self):
        """Processes structures from files into PyTorch Geometric Data."""
        # Preprocess PDB files
        if self.pdb_transform:
            self.transform_pdbs()

        idx = 0
        # Chunk dataset for parallel processing
        chunk_size = 128

        def divide_chunks(l: List[str], n: int = 2) -> Generator:
            for i in range(0, len(l), n):
                yield l[i : i + n]

        # chunks = list(divide_chunks(self.structures, chunk_size))
        chunks: List[int] = list(
            divide_chunks(list(self.examples.keys()), chunk_size)
        )

        for chunk in tqdm(chunks):
            pdbs = [self.examples[idx] for idx in chunk]
            # Get chain selections
            if self.chain_selection_map is not None:
                chain_selections = [
                    self.chain_selection_map[idx] for idx in chunk
                ]
            else:
                chain_selections = ["all"] * len(chunk)

            # Create graph objects
            file_names = [f"{self.raw_dir}/{pdb}.pdb" for pdb in pdbs]

            graphs = construct_graphs_mp(
                path_it=file_names,
                config=self.config,
                chain_selections=chain_selections,
                return_dict=False,
            )
            if self.graph_transformation_funcs is not None:
                graphs = [self.transform_graphein_graphs(g) for g in graphs]

            # Convert to PyTorch Geometric Data
            graphs = [self.graph_format_convertor(g) for g in graphs]

            # Assign labels
            if self.graph_label_map:
                labels = [self.graph_label_map[idx] for idx in chunk]
                for i, _ in enumerate(chunk):
                    graphs[i].graph_y = labels[i]
            if self.node_label_map:
                labels = [self.node_label_map[idx] for idx in chunk]
                for i, _ in enumerate(chunk):
                    graphs[i].graph_y = labels[i]

            data_list = graphs

            del graphs

            if self.pre_filter is not None:
                data_list = [g for g in data_list if self.pre_filter(g)]

            if self.pre_transform is not None:
                data_list = [self.pre_transform(data) for data in data_list]

            for i, (pdb, chain) in enumerate(zip(pdbs, chain_selections)):
                if self.chain_selection_map is None:
                    torch.save(
                        data_list[i],
                        os.path.join(self.processed_dir, f"{pdb}.pt"),
                    )
                else:
                    torch.save(
                        data_list[i],
                        os.path.join(self.processed_dir, f"{pdb}_{chain}.pt"),
                    )

    def get(self, idx: int):
        """
        Returns PyTorch Geometric Data object for a given index.

        :param idx: Index to retrieve.
        :type idx: int
        :return: PyTorch Geometric Data object.
        """
        if self.chain_selection_map is not None:
            return torch.load(
                os.path.join(
                    self.processed_dir,
                    f"{self.structures[idx]}_\
                        {self.chain_selection_map[idx]}.pt",
                ),
                weights_only=False,
            )
        else:
            return torch.load(
                os.path.join(self.processed_dir, f"{self.structures[idx]}.pt"),
                weights_only=False,
            )


class ProteinGraphListDataset(InMemoryDataset):
    def __init__(
        self, root: str, data_list: List[Data], name: str, transform=None
    ):
        """Creates a dataset from a list of PyTorch Geometric Data objects.

        :param root: Root directory where the dataset is stored.
        :type root: str
        :param data_list: List of protein graphs as PyTorch Geometric Data
            objects.
        :type data_list: List[Data]
        :param name: Name of dataset. Data will be saved as ``data_{name}.pt``.
        :type name: str
        :param transform: A function/transform that takes in an
            :obj:`torch_geometric.data.Data` object and returns a transformed
            version. The data object will be transformed before every access.
            (default: :obj:`None`)
        :type transform: Optional[Callable], optional
        """
        self.data_list = data_list
        self.name = name
        super().__init__(root, transform)
        self.data, self.slices = torch.load(
            self.processed_paths[0], weights_only=False
        )

    @property
    def processed_file_names(self):
        """The name of the files in the :obj:`self.processed_dir` folder that
        must be present in order to skip processing."""
        return f"data_{self.name}.pt"

    def process(self):
        """Saves data files to disk."""
        torch.save(self.collate(self.data_list), self.processed_paths[0])



"""Functions for computing atomic structure of proteins."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from typing import Any, Dict

import networkx as nx
import numpy as np
import pandas as pd
from loguru import logger as log

from graphein.protein.edges.distance import compute_distmat
from graphein.protein.resi_atoms import (
    BOND_LENGTHS,
    BOND_ORDERS,
    COVALENT_RADII,
    DEFAULT_BOND_STATE,
    RESIDUE_ATOM_BOND_STATE,
)

# TODO dealing with metals
# TODO There are other check and balances that can be implemented from here:
# https://www.daylight.com/meetings/mug01/Sayle/m4xbondage.html


def assign_bond_states_to_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    """
    Takes a ``PandasPDB`` atom DataFrame and assigns bond states to each atom
    based on:

        *Atomic Structures of all the Twenty Essential Amino Acids and a*
        *Tripeptide, with Bond Lengths as Sums of Atomic Covalent Radii*
        Heyrovska, 2008

    First, maps atoms to their standard bond states
    (:const:`~graphein.protein.resi_atoms.DEFAULT_BOND_STATE`). Second, maps
    non-standard bonds states
    (:const:`~graphein.protein.resi_atoms.RESIDUE_ATOM_BOND_STATE`). Fills
    ``NaNs`` with standard bond states.

    :param df: Pandas PDB DataFrame.
    :type df: pd.DataFrame
    :return: DataFrame with added ``atom_bond_state`` column.
    :rtype: pd.DataFrame
    """

    # Map atoms to their standard bond states
    naive_bond_states = pd.Series(df["atom_name"].map(DEFAULT_BOND_STATE))

    # Create series of bond states for the non-standard states
    ss = (
        pd.DataFrame(RESIDUE_ATOM_BOND_STATE)
        .unstack()
        .rename_axis(("residue_name", "atom_name"))
        .rename("atom_bond_state")
    )

    # Map non-standard states to the DataFrame based on the residue and atom
    # name
    df = df.join(ss, on=["residue_name", "atom_name"])

    # Fill the NaNs with the standard states
    df = df.fillna(value={"atom_bond_state": naive_bond_states})

    return df


def assign_covalent_radii_to_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    """
    Assigns covalent radius
    (:const:`~graphein.protein.resi_atoms.COVALENT_RADII`) to each atom based
    on its bond state. Adds a ``covalent_radius`` column. Using values from:

        *Atomic Structures of all the Twenty Essential Amino Acids and a*
        *Tripeptide, with Bond Lengths as Sums of Atomic Covalent Radii*
        Heyrovska, 2008

    :param df: Pandas PDB DataFrame with a ``bond_states_column``.
    :type df: pd.DataFrame
    :return: Pandas PDB DataFrame with added ``covalent_radius`` column.
    :rtype: pd.DataFrame
    """
    # Assign covalent radius to each atom
    df["covalent_radius"] = df["atom_bond_state"].map(COVALENT_RADII)

    return df


def add_atomic_edges(G: nx.Graph, tolerance: float = 0.56) -> nx.Graph:
    """
    Computes covalent edges based on atomic distances. Covalent radii are
    assigned to each atom based on its bond assign_bond_states_to_dataframe.
    The distance matrix is then thresholded to entries less than this distance
    plus some tolerance to create an adjacency matrix. This adjacency matrix is
    then parsed into an edge list and covalent edges added

    :param G: Atomic graph (nodes correspond to atoms) to populate with atomic
        bonds as edges
    :type G: nx.Graph
    :param tolerance: Tolerance for atomic distance. Default is ``0.56``
        Angstroms. Commonly used values are: ``0.4, 0.45, 0.56``
    :type tolerance: float
    :return: Atomic graph with edges between bonded atoms added
    :rtype: nx.Graph
    """
    dist_mat = compute_distmat(G.graph["pdb_df"])

    # We assign bond states to the dataframe, and then map these to covalent
    # radii
    G.graph["pdb_df"] = assign_bond_states_to_dataframe(G.graph["pdb_df"])
    G.graph["pdb_df"] = assign_covalent_radii_to_dataframe(G.graph["pdb_df"])

    # Create a covalent 'distance' matrix by adding the radius arrays with its
    # transpose
    covalent_radius_distance_matrix = np.add(
        np.array(G.graph["pdb_df"]["covalent_radius"]).reshape(-1, 1),
        np.array(G.graph["pdb_df"]["covalent_radius"]).reshape(1, -1),
    )

    # Add the tolerance
    covalent_radius_distance_matrix = (
        covalent_radius_distance_matrix + tolerance
    )

    # Threshold Distance Matrix to entries where the eucl distance is less than
    # the covalent radius plus tolerance and larger than 0.4
    dist_mat = dist_mat[dist_mat > 0.4]
    t_distmat = dist_mat[dist_mat < covalent_radius_distance_matrix]

    # Store atomic adjacency matrix in graph
    G.graph["atomic_adj_mat"] = np.nan_to_num(t_distmat)

    # Get node IDs from non NaN entries in the thresholded distance matrix and
    # add the edge to the graph
    inds = zip(*np.where(~np.isnan(t_distmat)))
    for i in inds:
        length = t_distmat[i[0]][i[1]]
        node_1 = G.graph["pdb_df"]["node_id"][i[0]]
        node_2 = G.graph["pdb_df"]["node_id"][i[1]]
        chain_1 = G.graph["pdb_df"]["chain_id"][i[0]]
        chain_2 = G.graph["pdb_df"]["chain_id"][i[1]]

        # Check nodes are in graph
        if not (G.has_node(node_1) and G.has_node(node_2)):
            continue

        # Check atoms are in the same chain
        if chain_1 != chain_2:
            continue

        if G.has_edge(node_1, node_2):
            G.edges[node_1, node_2]["kind"].add("covalent")
            G.edges[node_1, node_2]["bond_length"] = length
        else:
            G.add_edge(node_1, node_2, kind={"covalent"}, bond_length=length)

    # Todo checking degree against MAX_NEIGHBOURS

    return G


def add_ring_status(G: nx.Graph) -> nx.Graph:
    """
    Identifies rings in the atomic graph. Assigns the edge attribute ``"RING"``
    to edges in the ring. We do not distinguish between aromatic and
    non-aromatic rings. Functions by identifying all cycles in the graph.

    :param G: Atom-level protein structure graph to add ring edge types to.
    :type G: nx.Graph
    :return: Atom-level protein structure graph with added ``"RING"`` edge
        attribute.
    :rtype: nx.Graph
    """
    cycles = nx.cycle_basis(
        G
    )  # Produces a list of lists containing nodes in each cycle
    # Iterate over cycles, check for an edge between the nodes
    # if there is one, add a "RING" attribute
    for cycle in cycles:
        [
            G.edges[x, y]["kind"].add("RING")
            for i, x in enumerate(cycle)
            for j, y in enumerate(cycle)
            if G.has_edge(x, y)
            if i != j
        ]

    return G


def add_bond_order(G: nx.Graph) -> nx.Graph:
    """
    Assign bond orders to the covalent bond edges between atoms on the basis of
    bond length. Values are taken from:

        *Automatic Assignment of Chemical Connectivity to Organic Molecules in*
        *the Cambridge Structural Database.*
        Jon C. Baber and Edward E. Hodgkin*

    :param G: Atomic-level protein graph with covalent edges.
    :type G: nx.Graph
    :return: Atomic-level protein graph with covalent edges annotated with
        putative bond order.
    :rtype: mx.Graph
    """
    for u, v, a in G.edges(data=True):
        atom_a = G.nodes[u]["element_symbol"]
        atom_b = G.nodes[v]["element_symbol"]

        # Assign bonds with hydrogens to 1
        if atom_a == "H" or atom_b == "H":
            G.edges[u, v]["kind"].add("SINGLE")
        # If not, we need to identify the bond type from the bond length
        else:
            query = f"{atom_a}-{atom_b}"
            # We need this try block as the dictionary keys may be X-Y, whereas
            # the query we construct may be Y-X
            try:
                identify_bond_type_from_mapping(G, u, v, a, query)
            except KeyError:
                query = f"{atom_b}-{atom_a}"
                try:
                    identify_bond_type_from_mapping(G, u, v, a, query)
                except KeyError:
                    log.debug(
                        f"Could not identify bond type for {query}. Adding a \
                            single bond."
                    )
                    G.edges[u, v]["kind"].add("SINGLE")

    return G


def identify_bond_type_from_mapping(
    G: nx.Graph, u: str, v: str, a: Dict[str, Any], query: str
):
    """
    Compares the bond length between two atoms in the graph, and the relevant
    experimental value by performing a lookup against the watershed values in:

        *Automatic Assignment of Chemical Connectivity to Organic Molecules in*
        *the Cambridge Structural Database.*
        Jon C. Baber and Edward E. Hodgkin*

    Bond orders are assigned in the order ``triple`` < ``double`` < ``single``
    (e.g. if a bond is shorter than the triple bond watershed (``w_dt``) then
    it is assigned as a triple bond. Similarly, if a bond is longer than this
    but shorter than the double bond watershed (``w_sd``), it is assigned double
    bond status.

    :param G: ``nx.Graph`` of atom-protein structure with atomic edges added.
    :type G: nx.Graph
    :param u: Node 1 in edge.
    :type u: str
    :param v: Node 2 in edge.
    :type v: str
    :param a: edge data
    :type a: Dict[str, Any]
    :param query: ``"ELEMENTX-ELEMENTY"`` to perform lookup with
        (E.g. ``"C-O"``,``"N-N"``)
    :type query: str
    :return: Graph with atomic edge bond order assigned
    :rtype: nx.Graph
    """
    # Perform lookup of allowable bond orders for the given atom pair
    allowable_order = BOND_ORDERS[query]
    # If max double, compare the length to the double watershed distance, w_sd,
    # else assign single
    if len(allowable_order) == 2:
        if a["bond_length"] < BOND_LENGTHS[query]["w_sd"]:
            G.edges[u, v]["kind"].add("DOUBLE")
        else:
            G.edges[u, v]["kind"].add("SINGLE")
    else:
        # If max triple, compare the length to the triple watershed distance,
        # w_dt, then double, else assign single
        if a["bond_length"] < BOND_LENGTHS[query]["w_dt"]:
            G.edges[u, v]["kind"].add("TRIPLE")
        elif a["bond_length"] < BOND_LENGTHS[query]["w_sd"]:
            G.edges[u, v]["kind"].add("DOUBLE")
        else:
            G.edges[u, v]["kind"].add("SINGLE")
    return G


# The codeblock below was used in an initial pass at solving the bond order
# assignment problem based on hybridisation state.
# We instead use a simpler method of construction based on bond lengths,
# but I am loathe to remove this code as it may prove useful later
"""
def cosinus(x0, x1, x2):
    e0 = x0 - x1
    e1 = x2 - x1
    e0 = e0 / np.linalg.norm(e0)
    e1 = e1 / np.linalg.norm(e1)
    cosinus = np.dot(e0, e1)
    angle = np.arccos(cosinus)

    return 180 - np.degrees(angle)


def dihedral(x0, x1, x2, x3):
    b0 = -1.0 * (x1 - x0)
    b1 = x2 - x1
    b2 = x3 - x2

    b0xb1 = np.cross(b0, b1)
    b1xb2 = np.cross(b2, b1)

    b0xb1_x_b1xb2 = np.cross(b0xb1, b1xb2)

    y = np.dot(b0xb1_x_b1xb2, b1) * (1.0 / np.linalg.norm(b1))
    x = np.dot(b0xb1, b1xb2)

    grad = 180 - np.degrees(np.arctan2(y, x))
    return grad


def assign_bond_orders(G: nx.Graph) -> nx.Graph:

    bond_angles: Dict[str, float] = {}
    for n, d in G.nodes(data=True):
        neighbours = list(G.neighbors(n))

        if len(neighbours) == 1:
            G.edges[n, neighbours[0]]["kind"].add("SINGLE")
            bond_angles[n] = 0.0
        elif len(neighbours) == 2:
            cos_angle = cosinus(
                G.nodes[n]["coords"],
                G.nodes[neighbours[0]]["coords"],
                G.nodes[neighbours[1]]["coords"],
            )
            bond_angles[n] = cos_angle
        elif len(neighbours) == 3:
            dihed = dihedral(
                G.nodes[n]["coords"],
                G.nodes[neighbours[0]]["coords"],
                G.nodes[neighbours[1]]["coords"],
                G.nodes[neighbours[2]]["coords"],
            )
            bond_angles[n] = dihed

    print(bond_angles)

    # Assign Bond angles to dataframe
    G.graph["pdb_df"]["bond_angles"] = G.graph["pdb_df"]["node_id"].map(
        bond_angles
    )
    print(G.graph["pdb_df"].to_string())

    # Assign Hybridisation state from Bond Angles
    hybridisation_state = {
        n: "sp"
        if d > 155
        else "sp2"
        if d > 115
        else "sp3"
        if d <= 115
        else "UNK"
        for n, d in bond_angles.items()
    }
    G.graph["pdb_df"]["bond_angles"] = G.graph["pdb_df"]["node_id"].map(
        hybridisation_state
    )

    return G
"""



"""
Author: Eric J. Ma, Arian Jamasb
Purpose: This is a set of utility variables and functions related to small molecules that can be used
across the Graphein project.

These include various collections of standard atom types used molecule-focussed ML
"""

# Graphein
# Author: Eric J. Ma, Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

from typing import Dict, List

from loguru import logger

from graphein.utils.dependencies import import_message

try:
    import rdkit.Chem as Chem
except (ImportError, ModuleNotFoundError):
    logger.warning(
        import_message(__name__, "rdkit", "rdkit", True, extras=True)
    )

BASE_ATOMS: List[str] = [
    "C",
    "H",
    "O",
    "N",
    "F",
    "P",
    "S",
    "Cl",
    "Br",
    "I",
    "B",
]
"""Vocabulary of 11 standard atom types."""

EXTENDED_ATOMS = [
    "C",
    "N",
    "O",
    "S",
    "F",
    "Si",
    "P",
    "Cl",
    "Br",
    "Mg",
    "Na",
    "Ca",
    "Fe",
    "As",
    "Al",
    "I",
    "B",
    "V",
    "K",
    "Tl",
    "Yb",
    "Sb",
    "Sn",
    "Ag",
    "Pd",
    "Co",
    "Se",
    "Ti",
    "Zn",
    "H",
    "Li",
    "Ge",
    "Cu",
    "Au",
    "Ni",
    "Cd",
    "In",
    "Mn",
    "Zr",
    "Cr",
    "Pt",
    "Hg",
    "Pb",
    "Unknown",
]
"""Vocabulary of additional atom types."""

ALLOWED_DEGREES: List[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
"""Vocabulary of allowed atom degrees."""

ALLOWED_VALENCES: List[int] = [0, 1, 2, 3, 4, 5, 6]
"""Vocabulary of allowed atom valences."""

ALLOWED_HYBRIDIZATIONS: List[Chem.rdchem.HybridizationType] = [
    Chem.rdchem.HybridizationType.SP,
    Chem.rdchem.HybridizationType.SP2,
    Chem.rdchem.HybridizationType.SP3,
    Chem.rdchem.HybridizationType.SP3D,
    Chem.rdchem.HybridizationType.SP3D2,
]
"""Vocabulary of allowed hybridizations."""

ALLOWED_NUM_H: List[int] = [0, 1, 2, 3, 4]
"""Vocabulary of allowed number of Hydrogens."""

ALLOWED_BOND_TYPES: List[Chem.rdchem.BondType] = [
    Chem.rdchem.BondType.SINGLE,
    Chem.rdchem.BondType.DOUBLE,
    Chem.rdchem.BondType.TRIPLE,
    Chem.rdchem.BondType.AROMATIC,
]
"""Vocabulary of allowed bondtypes."""

ALLOWED_BOND_TYPE_TO_CHANNEL: Dict[Chem.rdchem.BondType, int] = {
    Chem.rdchem.BondType.SINGLE: 0,
    Chem.rdchem.BondType.DOUBLE: 1,
    Chem.rdchem.BondType.TRIPLE: 2,
    Chem.rdchem.BondType.AROMATIC: 3,
}
"""Mapping of bondtypes to integer values."""


ALL_BOND_TYPES: List[Chem.rdchem.BondType] = [
    Chem.rdchem.BondType.AROMATIC,
    Chem.rdchem.BondType.DATIVE,
    Chem.rdchem.BondType.DATIVEL,
    Chem.rdchem.BondType.DATIVER,
    Chem.rdchem.BondType.DOUBLE,
    Chem.rdchem.BondType.FIVEANDAHALF,
    Chem.rdchem.BondType.FOURANDAHALF,
    Chem.rdchem.BondType.HEXTUPLE,
    Chem.rdchem.BondType.HYDROGEN,
    Chem.rdchem.BondType.IONIC,
    Chem.rdchem.BondType.ONEANDAHALF,
    Chem.rdchem.BondType.OTHER,
    Chem.rdchem.BondType.QUADRUPLE,
    Chem.rdchem.BondType.QUINTUPLE,
    Chem.rdchem.BondType.SINGLE,
    Chem.rdchem.BondType.THREEANDAHALF,
    Chem.rdchem.BondType.THREECENTER,
    Chem.rdchem.BondType.TRIPLE,
    Chem.rdchem.BondType.TWOANDAHALF,
    Chem.rdchem.BondType.UNSPECIFIED,
    Chem.rdchem.BondType.ZERO,
]
"""Vocabulary of all RDkit BondTypes."""


ALL_BOND_TYPES_TO_CHANNEL: Dict[Chem.rdchem.BondType, int] = {
    Chem.rdchem.BondType.AROMATIC: 0,
    Chem.rdchem.BondType.DATIVE: 1,
    Chem.rdchem.BondType.DATIVEL: 2,
    Chem.rdchem.BondType.DATIVER: 3,
    Chem.rdchem.BondType.DOUBLE: 4,
    Chem.rdchem.BondType.FIVEANDAHALF: 5,
    Chem.rdchem.BondType.FOURANDAHALF: 6,
    Chem.rdchem.BondType.HEXTUPLE: 7,
    Chem.rdchem.BondType.HYDROGEN: 8,
    Chem.rdchem.BondType.IONIC: 9,
    Chem.rdchem.BondType.ONEANDAHALF: 10,
    Chem.rdchem.BondType.OTHER: 11,
    Chem.rdchem.BondType.QUADRUPLE: 12,
    Chem.rdchem.BondType.QUINTUPLE: 13,
    Chem.rdchem.BondType.SINGLE: 14,
    Chem.rdchem.BondType.THREEANDAHALF: 15,
    Chem.rdchem.BondType.THREECENTER: 16,
    Chem.rdchem.BondType.TRIPLE: 17,
    Chem.rdchem.BondType.TWOANDAHALF: 18,
    Chem.rdchem.BondType.UNSPECIFIED: 19,
    Chem.rdchem.BondType.ZERO: 20,
}
"""Vocabulary of all RDkit BondTypes mapped to integer values."""

ALL_STEREO_TYPES: List[Chem.rdchem.BondStereo] = [
    Chem.rdchem.BondStereo.STEREOANY,
    Chem.rdchem.BondStereo.STEREOCIS,
    Chem.rdchem.BondStereo.STEREOE,
    Chem.rdchem.BondStereo.STEREONONE,
    Chem.rdchem.BondStereo.STEREOTRANS,
    Chem.rdchem.BondStereo.STEREOZ,
]
"""Vocabulary of all RDKit bond stereo types."""

ALL_STEREO_TO_CHANNEL: Dict[Chem.rdchem.BondStereo, int] = {
    Chem.rdchem.BondStereo.STEREOANY: 0,
    Chem.rdchem.BondStereo.STEREOCIS: 1,
    Chem.rdchem.BondStereo.STEREOE: 2,
    Chem.rdchem.BondStereo.STEREONONE: 3,
    Chem.rdchem.BondStereo.STEREOTRANS: 4,
    Chem.rdchem.BondStereo.STEREOZ: 5,
}
"""Vocabulary of all RDKit bond stereo types mapped to integer values."""

CHIRAL_TYPE: List[Chem.rdchem.ChiralType] = [
    Chem.rdchem.ChiralType.CHI_OTHER,
    Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW,
    Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
    Chem.rdchem.ChiralType.CHI_UNSPECIFIED,
]
"""Vocabulary of all RDKit chiral types."""

CHIRAL_TYPE_TO_CHANNEL: Dict[Chem.rdchem.ChiralType, int] = {
    Chem.rdchem.ChiralType.CHI_OTHER: 0,
    Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW: 1,
    Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW: 2,
    Chem.rdchem.ChiralType.CHI_UNSPECIFIED: 3,
}
"""Vocabulary of all RDKit chiral types mapped to integer values."""


RDKIT_MOL_DESCRIPTORS: List[str] = [
    "MaxEStateIndex",
    "MinEStateIndex",
    "MaxAbsEStateIndex",
    "MinAbsEStateIndex",
    "qed",
    "MolWt",
    "HeavyAtomMolWt",
    "ExactMolWt",
    "NumValenceElectrons",
    "NumRadicalElectrons",
    "MaxPartialCharge",
    "MinPartialCharge",
    "MaxAbsPartialCharge",
    "MinAbsPartialCharge",
    "FpDensityMorgan1",
    "FpDensityMorgan2",
    "FpDensityMorgan3",
    "BCUT2D_MWHI",
    "BCUT2D_MWLOW",
    "BCUT2D_CHGHI",
    "BCUT2D_CHGLO",
    "BCUT2D_LOGPHI",
    "BCUT2D_LOGPLOW",
    "BCUT2D_MRHI",
    "BCUT2D_MRLOW",
    "BalabanJ",
    "BertzCT",
    "Chi0",
    "Chi0n",
    "Chi0v",
    "Chi1",
    "Chi1n",
    "Chi1v",
    "Chi2n",
    "Chi2v",
    "Chi3n",
    "Chi3v",
    "Chi4n",
    "Chi4v",
    "HallKierAlpha",
    "Ipc",
    "Kappa1",
    "Kappa2",
    "Kappa3",
    "LabuteASA",
    "PEOE_VSA1",
    "PEOE_VSA10",
    "PEOE_VSA11",
    "PEOE_VSA12",
    "PEOE_VSA13",
    "PEOE_VSA14",
    "PEOE_VSA2",
    "PEOE_VSA3",
    "PEOE_VSA4",
    "PEOE_VSA5",
    "PEOE_VSA6",
    "PEOE_VSA7",
    "PEOE_VSA8",
    "PEOE_VSA9",
    "SMR_VSA1",
    "SMR_VSA10",
    "SMR_VSA2",
    "SMR_VSA3",
    "SMR_VSA4",
    "SMR_VSA5",
    "SMR_VSA6",
    "SMR_VSA7",
    "SMR_VSA8",
    "SMR_VSA9",
    "SlogP_VSA1",
    "SlogP_VSA10",
    "SlogP_VSA11",
    "SlogP_VSA12",
    "SlogP_VSA2",
    "SlogP_VSA3",
    "SlogP_VSA4",
    "SlogP_VSA5",
    "SlogP_VSA6",
    "SlogP_VSA7",
    "SlogP_VSA8",
    "SlogP_VSA9",
    "TPSA",
    "EState_VSA1",
    "EState_VSA10",
    "EState_VSA11",
    "EState_VSA2",
    "EState_VSA3",
    "EState_VSA4",
    "EState_VSA5",
    "EState_VSA6",
    "EState_VSA7",
    "EState_VSA8",
    "EState_VSA9",
    "VSA_EState1",
    "VSA_EState10",
    "VSA_EState2",
    "VSA_EState3",
    "VSA_EState4",
    "VSA_EState5",
    "VSA_EState6",
    "VSA_EState7",
    "VSA_EState8",
    "VSA_EState9",
    "FractionCSP3",
    "HeavyAtomCount",
    "NHOHCount",
    "NOCount",
    "NumAliphaticCarbocycles",
    "NumAliphaticHeterocycles",
    "NumAliphaticRings",
    "NumAromaticCarbocycles",
    "NumAromaticHeterocycles",
    "NumAromaticRings",
    "NumHAcceptors",
    "NumHDonors",
    "NumHeteroatoms",
    "NumRotatableBonds",
    "NumSaturatedCarbocycles",
    "NumSaturatedHeterocycles",
    "NumSaturatedRings",
    "RingCount",
    "MolLogP",
    "MolMR",
    "fr_Al_COO",
    "fr_Al_OH",
    "fr_Al_OH_noTert",
    "fr_ArN",
    "fr_Ar_COO",
    "fr_Ar_N",
    "fr_Ar_NH",
    "fr_Ar_OH",
    "fr_COO",
    "fr_COO2",
    "fr_C_O",
    "fr_C_O_noCOO",
    "fr_C_S",
    "fr_HOCCN",
    "fr_Imine",
    "fr_NH0",
    "fr_NH1",
    "fr_NH2",
    "fr_N_O",
    "fr_Ndealkylation1",
    "fr_Ndealkylation2",
    "fr_Nhpyrrole",
    "fr_SH",
    "fr_aldehyde",
    "fr_alkyl_carbamate",
    "fr_alkyl_halide",
    "fr_allylic_oxid",
    "fr_amide",
    "fr_amidine",
    "fr_aniline",
    "fr_aryl_methyl",
    "fr_azide",
    "fr_azo",
    "fr_barbitur",
    "fr_benzene",
    "fr_benzodiazepine",
    "fr_bicyclic",
    "fr_diazo",
    "fr_dihydropyridine",
    "fr_epoxide",
    "fr_ester",
    "fr_ether",
    "fr_furan",
    "fr_guanido",
    "fr_halogen",
    "fr_hdrzine",
    "fr_hdrzone",
    "fr_imidazole",
    "fr_imide",
    "fr_isocyan",
    "fr_isothiocyan",
    "fr_ketone",
    "fr_ketone_Topliss",
    "fr_lactam",
    "fr_lactone",
    "fr_methoxy",
    "fr_morpholine",
    "fr_nitrile",
    "fr_nitro",
    "fr_nitro_arom",
    "fr_nitro_arom_nonortho",
    "fr_nitroso",
    "fr_oxazole",
    "fr_oxime",
    "fr_para_hydroxylation",
    "fr_phenol",
    "fr_phenol_noOrthoHbond",
    "fr_phos_acid",
    "fr_phos_ester",
    "fr_piperdine",
    "fr_piperzine",
    "fr_priamide",
    "fr_prisulfonamd",
    "fr_pyridine",
    "fr_quatN",
    "fr_sulfide",
    "fr_sulfonamd",
    "fr_sulfone",
    "fr_term_acetylene",
    "fr_tetrazole",
    "fr_thiazole",
    "fr_thiocyan",
    "fr_thiophene",
    "fr_unbrch_alkane",
    "fr_urea",
]
"""Vocabulary of easy-to-compute RDKit molecule descriptors"""



"""Functions for featurising Small Molecule Graphs."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Yuanqi Du
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd

from graphein.molecule.atoms import BASE_ATOMS
from graphein.utils.utils import onek_encoding_unk


def atom_type_one_hot(
    n,
    d: Dict[str, Any],
    return_array: bool = True,
    allowable_set: Optional[List[str]] = None,
) -> np.ndarray:
    """Adds a one-hot encoding of atom types as a node attribute.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data.
    :type d: Dict[str, Any]
    :param return_array: If ``True``, returns a numpy ``np.ndarray`` of one-hot encoding, otherwise returns a ``pd.Series``. Default is ``True``.
    :type return_array: bool
    :param allowable_set: Specifies vocabulary of amino acids. Default is ``None`` (which uses `graphein.molecule.atoms.BASE_ATOMS`).
    :return: One-hot encoding of amino acid types.
    :rtype: Union[pd.Series, np.ndarray]
    """

    if allowable_set is None:
        allowable_set = BASE_ATOMS

    features = onek_encoding_unk(d["element"], allowable_set)

    if return_array:
        features = np.array(features).astype(int)
    else:
        features = pd.Series(features).astype(int)
        features.index = allowable_set

    d["atom_type_one_hot"] = features
    return features


def degree(n: str, d: Dict[str, Any]) -> int:
    """Adds the degree of the node to the node data.

    N.B. this is the degree as defined by RDKit rather than the 'true' degree of the node in the graph. For the latter, use nx.degree()

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Degree of the atom.
    :rtype: int
    """
    degree = d["rdmol_atom"].GetDegree()
    d["degree"] = degree
    return degree


def degree_one_hot(n: str, d, allowable_set: List[int]):
    deg = degree(n, d)
    deg = onek_encoding_unk(deg, allowable_set)
    d["degree"] = deg
    return deg


def total_degree(n: str, d: Dict[str, Any]) -> int:
    """Adds the total degree of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data.
    :type d: Dict[str, Any]
    :return: Total degree of the atom.
    :rtype: int
    """
    total_degree = d["rdmol_atom"].GetTotalDegree()
    d["total_degree"] = total_degree
    return total_degree


def total_valence(n: str, d: Dict[str, Any]) -> int:
    """Adds the total valence of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data.
    :type d: Dict[str, Any]
    :return: Total valence of the atom.
    :rtype: int
    """
    total_valence = d["rdmol_atom"].GetTotalValence()
    d["total_valence"] = total_valence
    return total_valence


def explicit_valence(n: str, d: Dict[str, Any]) -> int:
    """Adds explicit valence of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Explicit valence of the atom.
    :rtype: int
    """
    explicit_valence = d["rdmol_atom"].GetExplicitValence()
    d["explicit_valence"] = explicit_valence
    return explicit_valence


def implicit_valence(n: str, d: Dict[str, Any]) -> int:
    """Adds implicit valence of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Implicit valence of the atom.
    :rtype: int
    """
    implicit_valence = d["rdmol_atom"].GetImplicitValence()
    d["implicit_valence"] = implicit_valence
    return implicit_valence


def num_implicit_h(n: str, d: Dict[str, Any]) -> int:
    """Adds the number of implicit Hydrogens of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Number of implicit Hydrogens of the atom.
    :rtype: int
    """
    implicit_hs = d["rdmol_atom"].GetNumImplicitHs()
    d["num_implicit_h"] = implicit_hs
    return implicit_hs


def num_explicit_h(n: str, d: Dict[str, Any]) -> int:
    """Adds the number of explicit Hydrogens of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Number of explicit Hydrogens of the atom.
    :rtype: int
    """
    num_explicit_h = d["rdmol_atom"].GetNumExplicitHs()
    d["num_explicit_h"] = num_explicit_h
    return num_explicit_h


def total_num_h(n: str, d: Dict[str, Any]) -> int:
    """Adds the total number of Hydrogens of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Total number of Hydrogens of the atom.
    :rtype: int
    """
    total_num_h = d["rdmol_atom"].GetTotalNumHs()
    d["total_num_h"] = total_num_h
    return total_num_h


def num_radical_electrons(n: str, d: Dict[str, Any]) -> int:
    """Adds the number of radical electrons of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Number of radical electrons of the atom.
    :rtype: int
    """
    num_radical_electrons = d["rdmol_atom"].GetNumRadicalElectrons()
    d["num_radical_electrons"] = num_radical_electrons
    return num_radical_electrons


def formal_charge(n: str, d: Dict[str, Any]) -> int:
    """Adds the formal charge of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Formal charge of the atom.
    :rtype: int
    """
    formal_charge = d["rdmol_atom"].GetFormalCharge()
    d["formal_charge"] = formal_charge
    return formal_charge


def hybridization(
    n: str, d: Dict[str, Any]
) -> rdkit.Chem.rdchem.HybridizationType:
    """Adds the hybridization of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Hybridization of the atom.
    :rtype: rdkit.Chem.rdchem.HybridizationType
    """
    hybridization = d["rdmol_atom"].GetHybridization()
    d["hybridization"] = hybridization
    return hybridization


def is_aromatic(n: str, d: Dict[str, Any]) -> bool:
    """Adds indicator of aromaticity of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Indicator of aromaticity of the atom.
    :rtype: bool
    """
    aromatic = d["rdmol_atom"].GetIsAromatic()
    d["is_aromatic"] = aromatic
    return aromatic


def is_ring(n: str, d: Dict[str, Any]) -> bool:
    """Adds indicator of ring membership of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Indicator of ring membership of the atom.
    :rtype: bool
    """
    ring = d["rdmol_atom"].IsInRing()
    d["is_ring"] = ring
    return ring


def is_ring_size(n: str, d: Dict[str, Any], ring_size: int) -> bool:
    """Adds indicator of ring membership of size ``ring_size`` of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data.
    :type d: Dict[str, Any]
    :param ring_size: The size of the ring to look for.
    :type ring_size: int
    :return: Indicator of ring membership of size ``ring_size`` of the atom.
    :rtype: bool
    """
    ring = d["rdmol_atom"].IsInRingSize(ring_size)
    d[f"is_ring_{ring_size}"] = ring
    return ring


def is_isotope(n: str, d: Dict[str, Any]) -> int:
    """Adds indicator of whether or not the atom is an isotope to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Indicator of whether or not the atom is an isotope.
    :rtype: int
    """
    is_isotope = d["rdmol_atom"].GetIsotope()
    d["is_isotope"] = is_isotope
    return is_isotope


def atomic_mass(n: str, d: Dict[str, Any]) -> float:
    """Adds mass of the atom to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Mass of the atom.
    :rtype: float
    """
    mass = d["rdmol_atom"].GetMass()
    d["mass"] = mass
    return mass


def chiral_tag(n: str, d: Dict[str, Any]) -> rdkit.Chem.rdchem.ChiralType:
    """Adds indicator of atom chirality to the node data.

    :param n: Node name, this is unused and only included for compatibility with the other functions.
    :type n: str
    :param d: Node data
    :type d: Dict[str, Any]
    :return: Indicator of atom chirality.
    :rtype: rdkit.Chem.rdchem.ChiralType
    """
    tag = d["rdmol_atom"].GetChiralTag()
    d["chiral_tag"] = tag
    return tag



"""Utilities for using structure prediction models.

N.B. ESMFold functionality require torch & ESMFold to be installed.

.. code-block:: bash
    pip install "fair-esm[esmfold]"
    # OpenFold and its remaining dependency
    pip install 'dllogger @ git+https://github.com/NVIDIA/dllogger.git'
    pip install 'openfold @ git+https://github.com/aqlaboratory/openfold.git@4b41059694619831a7db195b7e0988fc4ff3a307'

"""

import os
import subprocess
from functools import lru_cache
from typing import Dict, List, Optional

import requests
from loguru import logger as log

from graphein.utils.dependencies import import_message, is_tool

try:
    import torch
except ImportError:
    message = import_message(
        "graphein.protein.folding_utils",
        package="torch",
        conda_channel="pytorch",
        pip_install=True,
    )
    log.warning(message)


try:
    import esm
except ImportError:
    message = import_message(
        "graphein.protein.folding_utils",
        package="fair-esm[esmfold]",
        pip_install=True,
    )
    log.warning(message)

try:
    import foldcomp
except ImportError:
    message = import_message(
        "graphein.protein.folding_utils",
        package="foldcomp",
        pip_install=True,
        extras=True,
    )
    log.warning(message)


@lru_cache
def _get_model(model: str = "v1") -> torch.nn.Module:
    """Loads the ESMFold model."""
    if model == "v1":
        model = esm.pretrained.esmfold_v1()
    elif model == "v0":
        model = esm.pretrained.esmfold_v0()

    model = model.eval().cuda()
    return model


def esmfold(sequence: str, out_path: str):
    """Fold a protein sequence using the ESMFold model.

    Multimer prediction can be done with chains separated by ``:``.

    :param sequence: Amino acid sequence in one-letter code.
    :type sequence: str
    :param out_path: Path to save the PDB file to.
    :type out_path: str
    """

    model = _get_model()

    with torch.no_grad():
        output = model.infer_pdb(sequence)

    with open(out_path, "w") as f:
        f.write(output)


def esmfold_fasta(
    fasta: str,
    out_dir: str,
    num_recycles: int = 4,
    max_tokens: Optional[int] = None,
    chunk_size: Optional[int] = None,
    cpu_only: bool = False,
    cpu_offload: bool = False,
):
    """Batch fold a fasta file with ESMFold.

    Multimer prediction can be done with chains separated by ``:``.

    :param fasta: Path to FASTA file
    :type fasta: str
    :param out_dir: Path to output directory
    :type out_dir: str
    :param num_recycles: Number of recycles to perform. Defaults to number used
        in training (4).
    :type num_recycles: int
    :param max_tokens: Maximum number of tokens per gpu forward-pass. This
        will group shorter sequences together for batched prediction. Lowering
        this can help with out of memory issues, if these occur on short
        sequences.
    :type max_tokens: int
    :param chunk_size: Chunks axial attention computation to reduce memory
        usage from O(L^2) to O(L). Equivalent to running a for loop over chunks
        of of each dimension. Lower values will result in lower memory usage at
        the cost of speed. Recommended values: ``128``, ``64``, ``32``.
        Default: ``None``.
    :type chunk_size: int
    :param cpu_only: CPU only
    :type cpu_only: bool
    :param cpu_offload: Enable CPU offloading
    :type cpu_offload: bool
    :raises FileNotFoundError: If fasta file not found.
    """
    is_tool("esm-fold", error=True)
    if not os.path.exists(fasta):
        raise FileNotFoundError(f"File {fasta} not found.")

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    command = f"esm-fold --fasta {fasta} "
    command += f"-o {out_dir} "
    command += f"--num-recycles {num_recycles} "
    if max_tokens is not None:
        command += f"--max-tokens-per-batch {max_tokens} "
    if chunk_size is not None:
        command += f"--chunk-size {chunk_size} "
    if cpu_only:
        command += "--cpu-only"
    if cpu_offload:
        command += "--cpu-offload"
    log.info(f"Running command: {command}")
    subprocess.run(command, shell=True)


def esm_embed_fasta(
    fasta: str,
    out_dir: str,
    model: str = "esm2_t33_650M_UR50D",
    max_tokens: Optional[int] = None,
    repr_layers: Optional[List[int]] = [0, 32, 33],
    include: Optional[List[str]] = ["mean", "per_tok"],
    truncation_seq_length: Optional[int] = None,
):
    """
    Batch embed a fasta file with ESM.


    Default parameters compute final-layer embedding from ESM-2.

    .. code-block:: bash

        python scripts/extract.py esm2_t33_650M_UR50D examples/data/some_proteins.fasta \
        examples/data/some_proteins_emb_esm2 --repr_layers 0 32 33 --include mean per_tok

    :param fasta: Path to FASTA file.
    :type fasta: str
    :param out_dir: Path to output directory.
    :type out_dir: str
    :param model: PyTorch model file OR name of pretrained model to
        download (see README for models) Defaults to ``esm2_t33_650M_UR50D``.
    :param repr_layers: layers indices from which to extract representations
                        (0 to num_layers, inclusive).
    :type repr_layers: List[int]
    :param max_tokens: Maximum number of tokens per gpu forward-pass.
    :type max_tokens: Optional[int]
    :param include: List of representations to include in the output.
        {mean,per_tok,bos,contacts} [{mean,per_tok,bos,contacts} ...]. Default
        is ``["mean", "per_tok"]``.
    :type include: List[str]
    :param truncation_seq_length: Truncate sequences longer than this value.
    :type truncation_seq_length: int
    """
    is_tool("esm-extract", error=True)
    if not os.path.exists(fasta):
        raise FileNotFoundError(f"File {fasta} not found.")

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    cmd = f"esm-extract {model} {fasta} {out_dir} "
    if repr_layers is not None:
        cmd += f"--repr_layers {' '.join([str(l) for l in repr_layers])} "
    if include is not None:
        cmd += f"--include {' '.join(include)} "
    if truncation_seq_length is not None:
        cmd += f"--truncation_seq_length {truncation_seq_length} "
    if max_tokens is not None:
        cmd += f"--toks_per_batch {max_tokens} "
    log.info(f"Running command: {cmd}")
    subprocess.run(cmd, shell=True)


def foldcompress_file(fname: str, anchor_residue_threshold: int = 25):
    """Compress a PDB file using the FoldCompress algorithm.

    :param fname: Path to PDB file.
    :type fname: str
    :param anchor_residue_threshold: Threshold for anchor residues. 25 should
        give a RMSD ~0.07A. A reset point every 200 residues will give a RMSD
        ~0.2A
    :type anchor_residue_threshold: int
    """
    if not os.path.exists(fname):
        raise FileNotFoundError(f"File {fname} not found.")
    if not fname.endswith(".pdb"):
        raise ValueError("Can only compress PDB files.")

    with open(fname, "r") as f:
        src = f.read()

    fcz = foldcomp.compress(fname, src, anchor_residue_threshold)

    stem = os.path.splitext(fname)[0]
    with open(f"{stem}.fcz", "wb") as f:
        f.write(fcz)


def foldcompress_database(
    db_name: str, dir_path: str, fnames: Optional[List[str]]
):
    """
    Compress all PDB files in the database using the FoldCompress algorithm.
    """
    is_tool("foldcomp", error=True)
    cmd = f"foldcomp compress {dir_path} {db_name} --db"
    log.info(f"Compressing {dir_path}. Running command: {cmd}")
    subprocess.run(cmd, shell=True)

    if fnames is not None:
        is_tool("mmseqs", error=True)
        log.info("Writing selection IDs")
        id_path = os.path.join(dir_path, f"{db_name}id_list.txt")
        with open(id_path, "w") as f:
            f.write("\n".join(fnames))

        cmd = f"mmseqs createsubdb --subdb-mode 0 --id-mode 1 {id_path} {db_name} {db_name}_selection"
        log.info(f"Extracing {len(fnames)} selections. Running command: {cmd}")
        subprocess.run(cmd, shell=True)


def esmfold_web(
    sequence: str, out_path: Optional[str] = None, version: int = 1
):
    """Fold a protein sequence using the ESMFold model from the ESMFold server
    at https://api.esmatlas.com/foldSequence/v1/pdb/.

    Parameters
    ----------
    sequence : str
        A protein sequence in one-letter code.
    out_path : str, optional
        Path to save the PDB file to. If `None`, the file is not saved.
        Defaults to `None`.
    version : int, optional
        The version of the ESMFold model to use. Defaults to `1`.
    """
    URL = f"https://api.esmatlas.com/foldSequence/v{version}/pdb/"

    headers: Dict[str, str] = {
        "Content-Type": "application/x-www-form-urlencoded",
    }

    cif = requests.post(URL, data=sequence, headers=headers).text
    # append header
    header = "\n".join(
        [f"data_{sequence}", "#", f"_entry.id\t{sequence}", "#\n"]
    )
    cif = header + cif
    if out_path is not None:
        with open(out_path, "w") as f:
            f.write(cif)



"""Functions for computing biochemical edges of graphs."""

# Graphein
# Author: Eric Ma, Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import itertools
from typing import Union

import networkx as nx
import numpy as np
import pandas as pd
from loguru import logger as log
from sklearn.metrics import pairwise_distances
from sklearn.neighbors import kneighbors_graph


def compute_distmat(coords: np.ndarray) -> np.ndarray:
    """
    Compute pairwise euclidean distances between every atom.

    Design choice: passed in a DataFrame to enable easier testing on
    dummy data.

    :param coords: pd.Dataframe containing molecule structure. Must contain
        columns ``["x_coord", "y_coord", "z_coord"]``.
    :type coords: pd.DataFrame
    :return: np.ndarray of euclidean distance matrix.
    :rtype: np.ndarray
    """
    return pairwise_distances(coords, metric="euclidean")


def get_interacting_atoms(angstroms: float, distmat: np.ndarray) -> np.ndarray:
    """Find the atoms that are within a particular radius of one another.

    :param angstroms: Radius in angstroms.
    :type angstroms: float
    :param distmat: Distance matrix.
    :type distmat: np.ndarray
    :returns: Array of interacting atoms
    :rtype: np.ndarray
    """
    return np.where(distmat <= angstroms)


def add_distance_threshold(G: nx.Graph, threshold: float = 5.0):
    """
    Adds edges to any nodes within a given distance of each other.

    :param G: molecule structure graph to add distance edges to
    :type G: nx.Graph
    :param threshold: Distance in angstroms, below which two nodes are
        connected.
    :type threshold: float
    :return: Graph with distance-based edges added
    """

    dist_mat = compute_distmat(G.graph["coords"])
    interacting_nodes = get_interacting_atoms(threshold, distmat=dist_mat)
    outgoing = [list(G.nodes())[i] for i in interacting_nodes[0]]
    incoming = [list(G.nodes())[i] for i in interacting_nodes[1]]
    interacting_nodes = list(zip(outgoing, incoming))

    log.info(
        f"Found: {len(interacting_nodes)} distance edges for radius {threshold}"
    )
    for n1, n2 in interacting_nodes:
        if G.has_edge(n1, n2):
            G.edges[n1, n2]["kind"].add("distance_threshold")
        else:
            G.add_edge(n1, n2, kind={"distance_threshold"})


def add_fully_connected_edges(
    G: nx.Graph,
):
    """
    Adds fully connected edges to nodes.

    :param G: Molecule structure graph to add distance edges to.
    :type G: nx.Graph
    """
    length = len(G.graph["coords"])

    for n1, n2 in itertools.product(G.nodes(), G.nodes()):
        if G.has_edge(n1, n2):
            G.edges[n1, n2]["kind"].add("fully_connected")
        else:
            G.add_edge(n1, n2, kind={"fully_connected"})


def add_k_nn_edges(
    G: nx.Graph,
    k: int = 1,
    mode: str = "connectivity",
    metric: str = "minkowski",
    p: int = 2,
    include_self: Union[bool, str] = False,
):
    """
    Adds edges to nodes based on K nearest neighbours.

    :param G: Molecule structure graph to add distance edges to.
    :type G: nx.Graph
    :param k: Number of neighbors for each sample.
    :type k: int
    :param mode: Type of returned matrix: ``"connectivity"`` will return the
        connectivity matrix with ones and zeros, and ``"distance"`` will return
        the distances between neighbors according to the given metric.
    :type mode: str
    :param metric: The distance metric used to calculate the k-Neighbors for
        each sample point. The DistanceMetric class gives a list of available
        metrics. The default distance is ``"euclidean"`` (``"minkowski"``
        metric with the ``p`` param equal to ``2``).
    :type metric: str
    :param p: Power parameter for the Minkowski metric. When ``p = 1``, this is
        equivalent to using ``manhattan_distance`` (l1), and
        ``euclidean_distance`` (l2) for ``p = 2``. For arbitrary ``p``,
        ``minkowski_distance`` (l_p) is used. Default is ``2`` (euclidean).
    :type p: int
    :param include_self: Whether or not to mark each sample as the first nearest
        neighbor to itself. If ``"auto"``, then ``True`` is used for
        ``mode="connectivity"`` and ``False`` for ``mode="distance"``. Default
        is ``False``.
    :type include_self: Union[bool, str]
    :return: Graph with knn-based edges added.
    :rtype: nx.Graph
    """
    dist_mat = compute_distmat(G.graph["coords"])

    nn = kneighbors_graph(
        X=dist_mat,
        n_neighbors=k,
        mode=mode,
        metric=metric,
        p=p,
        include_self=include_self,
    )

    # Create iterable of node indices
    outgoing = np.repeat(np.array(range(len(G.graph["coords"]))), k)
    outgoing = [list(G.nodes())[i] for i in outgoing]
    incoming = [list(G.nodes())[i] for i in nn.indices]
    interacting_nodes = list(zip(outgoing, incoming))
    log.info(f"Found: {len(interacting_nodes)} KNN edges")
    for n1, n2 in interacting_nodes:
        if G.has_edge(n1, n2):
            G.edges[n1, n2]["kind"].add(f"k_nn_{k}")
        else:
            G.add_edge(n1, n2, kind={f"k_nn_{k}"})



"""Featurization functions for graph edges."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Eric Ma, Charlie Harris
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import os
import subprocess
from pathlib import Path
from typing import Optional

import networkx as nx
import pandas as pd
from loguru import logger as log

from graphein.protein.utils import download_pdb

CONTACTS_FILE_SUFFIX = "_contacts.tsv"


def peptide_bonds(G: nx.Graph) -> nx.Graph:
    """
    Adds peptide backbone to residues in each chain.

    :param G: nx.Graph protein graph.
    :type G: nx.Graph
    :returns: nx.Graph protein graph with added peptide bonds.
    :rtype: nx.Graph
    """
    log.debug("Adding peptide bonds to graph")
    # Iterate over every chain
    for chain_id in G.graph["chain_ids"]:
        # Find chain residues
        chain_residues = [
            (n, v) for n, v in G.nodes(data=True) if v["chain_id"] == chain_id
        ]

        # Iterate over every residue in chain
        for i, residue in enumerate(chain_residues):
            # Checks not at chain terminus
            if i == len(chain_residues) - 1:
                continue
            # Asserts residues are on the same chain
            cond_1 = (
                residue[1]["chain_id"] == chain_residues[i + 1][1]["chain_id"]
            )
            # Asserts residue numbers are adjacent
            cond_2 = (
                abs(
                    residue[1]["residue_number"]
                    - chain_residues[i + 1][1]["residue_number"]
                )
                == 1
            )

            # If this checks out, we add a peptide bond
            if cond_1 and cond_2:
                # Adds "peptide_bond" between current residue and the next
                if G.has_edge(i, i + 1):
                    G.edges[i, i + 1]["kind"].add("peptide_bond")
                else:
                    G.add_edge(
                        residue[0],
                        chain_residues[i + 1][0],
                        kind={"peptide_bond"},
                    )
            else:
                continue
    return G


####################################
#                                  #
#     GetContacts Interactions     #
#                                  #
####################################


def get_contacts_df(config: GetContactsConfig, pdb_path: str) -> pd.DataFrame:
    """
    Reads GetContact File and returns it as a ``pd.DataFrame``.

    :param config: GetContactsConfig object
    :type config: GetContactsConfig
    :param pdb_name: Name of PDB file. Output contacts files are named:
        ``{pdb_name}_contacts.tsv``.
    :type pdb_name: str
    :return: DataFrame of parsed GetContacts output.
    :rtype: pd.DataFrame
    """
    if not config.contacts_dir:
        config.contacts_dir = Path("/tmp/")

    contacts_file = (
        config.contacts_dir / f"{Path(pdb_path).stem}{CONTACTS_FILE_SUFFIX}"
    )

    # Check for existence of GetContacts file
    if not os.path.isfile(contacts_file):
        log.info("GetContacts file not found. Running GetContacts...")
        run_get_contacts(config, file_name=pdb_path)

    contacts_df = read_contacts_file(config, contacts_file)

    # remove temp GetContacts file
    if config.contacts_dir == "/tmp/":
        os.remove(contacts_file)

    return contacts_df


def _validate_get_contacts_installation(config: GetContactsConfig) -> None:
    """Validate GetContacts installation."""
    get_contacts_script = (
        Path(config.get_contacts_path) / "get_static_contacts.py"
    )
    if not get_contacts_script.is_file():
        raise FileNotFoundError(
            f"GetContacts installation not found at {get_contacts_script}. "
            "Please install from: https://getcontacts.github.io"
        )


def _get_or_download_pdb(
    config: GetContactsConfig, pdb_id: Optional[str], file_name: Optional[str]
) -> Path:
    """Get PDB file path, downloading if necessary."""
    if file_name and Path(file_name).exists():
        return Path(file_name)

    pdb_path = config.pdb_dir / f"{pdb_id}{'.pdb'}"
    if not pdb_path.exists():
        log.debug(f"Downloading PDB file for {pdb_id}")
        downloaded_path = download_pdb(pdb_code=pdb_id, out_dir=config.pdb_dir)
        return Path(downloaded_path)

    return pdb_path


def run_get_contacts(
    config: GetContactsConfig,
    pdb_id: Optional[str] = None,
    file_name: Optional[str] = None,
):
    """
    Runs GetContacts on a protein structure. If no file_name is provided, a
    PDB file is downloaded for the pdb_id

    :param config: GetContactsConfig object containing GetContacts parameters
    :type config: graphein.protein.config.GetContactsConfig
    :param pdb_id: 4-character PDB accession code
    :type pdb_id: str, optional
    :param file_name: PDB_name file to use, if annotations to be retrieved from
        the PDB
    :type file_name: str, optional
    """
    _validate_get_contacts_installation(config)

    if not file_name and not pdb_id:
        raise ValueError("Either file_name or pdb_id must be provided")

    # Check for existence of pdb file. If not, download it.
    pdb_path = _get_or_download_pdb(config, pdb_id, file_name)
    output_path = config.contacts_dir / (pdb_path.stem + CONTACTS_FILE_SUFFIX)

    command = f"{config.get_contacts_path}/get_static_contacts.py "
    command += f"--structure {str(pdb_path)} "
    command += f"--output {str(output_path)} "
    command += "--itypes all"  # --sele "protein"'

    log.info(f"Running GetContacts: {command}")
    subprocess.run(command, shell=True, check=True)

    if not output_path.exists():
        raise FileNotFoundError(
            f"GetContacts failed to create output file: {output_path}"
        )

    log.info(f"Computed Contacts for: {pdb_id}")


def read_contacts_file(
    config: GetContactsConfig, contacts_file: str
) -> pd.DataFrame:
    """
    Parses GetContacts file to an edge list (pd.DataFrame).

    :param config: GetContactsConfig object
        (:ref:`~graphein.protein.config.GetContactsConfig`)
    :type config: GetContactsConfig
    :param contacts_file: file name of contacts file
    :type contacts_file: str
    :return: Pandas Dataframe of edge list
    :rtype: pd.DataFrame
    """
    if not Path(contacts_file).exists():
        raise FileNotFoundError(f"Contacts file not found: {contacts_file}")

    log.debug(f"Parsing GetContacts output file at: {contacts_file}")

    contacts = []

    # Extract every contact and residue types
    with open(contacts_file, "r") as f:
        # Skip header lines
        for _ in range(2):
            next(f)

        for line in f:
            contact = line.strip().split("\t")

            interaction_type = contact[1]
            res1, res2 = contact[2], contact[3]

            # Remove atom names if not using atom granularity
            if config.granularity != "atom":
                res1 = ":".join(res1.split(":")[:3])
                res2 = ":".join(res2.split(":")[:3])

            contacts.append([res1, res2, interaction_type])

    if not contacts:
        log.warning(f"No contacts found in file: {contacts_file}")
        return pd.DataFrame(columns=["res1", "res2", "interaction_type"])

    return pd.DataFrame(
        contacts, columns=["res1", "res2", "interaction_type"]
    ).drop_duplicates()


def _get_pdb_path_from_graph(G: nx.Graph) -> Path:
    """Extract PDB path from graph metadata."""
    if G.graph.get("path"):
        return Path(G.graph["path"])

    config = G.graph.get("config")
    if not config:
        raise KeyError("Graph missing required 'config' attribute")

    return (
        config.get_contacts_config.pdb_dir
        / f"{G.graph['name']}{'.pdb' if not str(G.graph['name']).endswith('.pdb') else ''}"
    )


def add_contacts_edge(G: nx.Graph, interaction_type: str) -> nx.Graph:
    """
    Adds specific interaction types to the protein graph.

    :param G: Networkx protein graph.
    :type G: nx.Graph
    :param interaction_type: Interaction type to be added.
    :type interaction_type: str
    :return G: nx.Graph with specified interaction-based edges added.
    :rtype: nx.Graph
    """
    log.debug(f"Adding {interaction_type} edges to graph")

    # Ensure contacts_df exists
    if "contacts_df" not in G.graph:
        log.info("No 'contacts_df' found in G.graph. Running GetContacts.")
        pdb_path = _get_pdb_path_from_graph(G)
        G.graph["contacts_df"] = get_contacts_df(
            G.graph["config"].get_contacts_config, str(pdb_path)
        )

    contacts = G.graph["contacts_df"]

    # Select specific interaction type
    interactions = contacts[contacts["interaction_type"] == interaction_type]

    for _, [res1, res2, interaction_type] in interactions.iterrows():
        # Check residues are actually in graph
        if not (G.has_node(res1) and G.has_node(res2)):
            continue

        if G.has_edge(res1, res2):
            G.edges[res1, res2]["kind"].add(interaction_type)
        else:
            G.add_edge(res1, res2, kind={interaction_type})

    return G


def _create_interaction_function(interaction_code: str, description: str):
    """Factory function to create interaction edge functions."""

    def interaction_func(G: nx.Graph) -> nx.Graph:
        """
        Adds {description} to protein structure graph.

        :param G: nx.Graph to add {description} to.
        :return: nx.Graph with {description} added.
        """
        return add_contacts_edge(G, interaction_code)

    interaction_func.__name__ = description.replace(" ", "_").replace("-", "_")
    interaction_func.__doc__ = (
        f"Adds {description} to protein structure graph."
    )
    return interaction_func


# Create all interaction functions
hydrogen_bond = _create_interaction_function("hb", "hydrogen bonds")
salt_bridge = _create_interaction_function("sb", "salt bridges")
pi_cation = _create_interaction_function("pc", "pi-cation interactions")
pi_stacking = _create_interaction_function("ps", "pi-stacking interactions")
t_stacking = _create_interaction_function("ts", "t-stacking interactions")
hydrophobic = _create_interaction_function("hp", "hydrophobic interactions")
van_der_waals = _create_interaction_function(
    "vdw", "van der Waals interactions"
)



"""Utilities for manipulating protein geometry."""

import copy

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from typing import List, Tuple, Union

from loguru import logger as log

from graphein.utils.dependencies import import_message

from ..resi_atoms import (
    CHI_ANGLES_ATOMS,
    IDEAL_BB_BOND_ANGLES,
    IDEAL_BB_BOND_LENGTHS,
)
from .angles import get_backbone_bond_angles, get_backbone_bond_lengths
from .representation import get_c_alpha, get_full_atom_coords
from .types import (
    AtomTensor,
    BackboneFrameTensor,
    BackboneTensor,
    CoordTensor,
    QuaternionTensor,
    RotationMatrix,
    RotationMatrixTensor,
)

try:
    import torch
    import torch.nn.functional as F
except ImportError:
    message = import_message(
        "graphein.protein.tensor.geometry",
        package="torch",
        conda_channel="pytorch",
        pip_install=True,
    )
    log.warning(message)

try:
    from torch_geometric.data import Batch
except ImportError:
    message = import_message(
        "graphein.protein.tensor.geometry",
        package="torch_geometric",
        conda_channel="pyg",
        pip_install=True,
    )
    log.warning(message)


def get_center(
    x: Union[AtomTensor, CoordTensor],
    ca_only: bool = True,
    fill_value: float = 1e-5,
) -> CoordTensor:
    """
    Returns the center of a protein.

    .. code-block:: python
        import torch

        x = torch.rand((10, 37, 3))
        get_center(x)


    .. seealso::

        :meth:`center_protein`


    :param x: Point Cloud to Center. Torch tensor of shape ``(Length , 3)`` or
        ``(Length, num atoms, 3)``.
    :param ca_only: If ``True``, only the C-alpha atoms will be used to compute
        the center. Only relevant with AtomTensor inputs. Default is ``False``.
    :type ca_only: bool
    :param fill_value: Value used to denote missing atoms. Default is )``1e-5)``.
    :type fill_value: float
    :return: Torch tensor of shape ``(N,D)`` -- Center of Point Cloud
    :rtype: Union[graphein.protein.tensor.types.AtomTensor, graphein.protein.tensor.types.CoordTensor]
    """
    if x.ndim != 3:
        return x.mean(dim=0)
    if ca_only:
        return get_c_alpha(x).mean(dim=0)

    x_flat, _, _ = get_full_atom_coords(x, fill_value=fill_value)
    return x_flat.mean(dim=0)


def center_protein(
    x: Union[AtomTensor, CoordTensor], ca_only: bool = True, fill_value=1e-5
) -> Union[AtomTensor, CoordTensor]:
    """
    Centers a protein in the coordinate system at the origin.

    .. seealso::

        :meth:`get_center`

    :param x: Point Cloud to Center. Torch tensor of shape ``(Length , 3)`` or
        ``(Length, num atoms, 3)``.
    :param ca_only: If ``True``, only the C-alpha atoms will be used to compute
        the center. Only relevant with AtomTensor inputs. Default is ``False``.
    :type ca_only: bool
    :param fill_value: Value used to denote missing atoms. Default is ``1e-5``.
    :type fill_value: float
    :return: Centered Point Cloud of same shape as input.
    :rtype: Union[graphein.protein.tensor.types.AtomTensor, graphein.protein.tensor.types.CoordTensor]
    """
    center = get_center(x, ca_only=ca_only, fill_value=fill_value)
    # Mask missing atoms
    fill_mask = torch.where(
        x == fill_value, torch.tensor(1.0), torch.tensor(0.0)
    ).bool()
    centered = x - center
    # Restore fill values
    centered[fill_mask] = fill_value
    return centered


def kabsch(
    A: Union[AtomTensor, CoordTensor],
    B: Union[AtomTensor, CoordTensor],
    ca_only: bool = True,
    residue_wise: bool = False,
    fill_value: float = 1e-5,
    return_transformed: bool = True,
    allow_reflections: bool = False,
) -> Union[
    CoordTensor,
    Tuple[BackboneFrameTensor, torch.Tensor],
    Tuple[RotationMatrix, torch.Tensor],
]:
    """
    Computes registration between two (2D or 3D) point clouds with known
    correspondences using Kabsch algorithm.

    Registration occurs in the zero centered coordinate system, and then
    must be transported back.

    .. see:: https://en.wikipedia.org/wiki/Kabsch_algorithm

    .. note::

        Based on implementation by Guillaume Bouvier (@bougui505):
        https://gist.github.com/bougui505/e392a371f5bab095a3673ea6f4976cc8

    :param A: Torch tensor of shape ``(N,D)`` -- Point Cloud to Align (source)
    :param B: Torch tensor of shape ``(N,D)`` -- Reference Point Cloud (target)
    :param ca_only: Whether to use only C-alpha atoms for alignment, defaults to
        ``True``. If ``False``, all atoms are used.
    :param fill_value: Value to fill in for missing atoms, defaults to ``1e-5``.
        Only relevant if ``ca_only=False``.
    :return: Torch tensor of shape ``(N,D)`` -- Aligned Point Cloud or Optimal
        rotation and translation. Rotation matrix is of shape ``(D,D)`` and for
        multiplication from the right.
    :rtype: Union[graphein.protein.tensor.types.CoordTensor, Tuple[torch.Tensor, torch.Tensor]]
    """

    if residue_wise:
        centroid_A = A[:, 1, :].view(-1, 1, 3)
        centroid_B = B[:, 1, :].view(-1, 1, 3)
    else:
        # Get center of mass
        centroid_A = get_center(A, ca_only=ca_only, fill_value=fill_value)
        centroid_B = get_center(B, ca_only=ca_only, fill_value=fill_value)

    AA = A - centroid_A
    BB = B - centroid_B

    # Covariance matrix
    H = AA.mT @ BB if residue_wise else AA.T @ BB
    U, _, Vt = torch.svd(H)

    if not allow_reflections:
        with torch.no_grad():
            if residue_wise:
                sign_flip = torch.det(U) * torch.det(Vt.mH) < 0
            else:
                det = torch.det(U) * torch.det(Vt.T)
                if det < 0.0:
                    print("Flipping!")
                    # Vt[-1, -1] *= -1
        if residue_wise:
            Vt_ = Vt.clone()
            Vt_[sign_flip, -1, :] *= -1

    R = Vt_ @ U.mT if residue_wise else Vt @ U.T

    t = centroid_B if residue_wise else centroid_B - R @ centroid_A
    if residue_wise:
        return (R @ AA.mT).mT + t if return_transformed else (R, t)
    else:
        return R.mm(A.T).T + t.T if return_transformed else (R, t)


# @torch.jit.script
def _sqrt_positive_part(x: torch.Tensor) -> torch.Tensor:
    """
    Returns ``torch.sqrt(torch.max(0, x))`` but with a zero subgradient where
    ``x`` is ``0``.
    """
    ret = torch.zeros_like(x)
    positive_mask = x > 0
    ret[positive_mask] = torch.sqrt(x[positive_mask])
    return ret


def matrix_to_quaternion(matrix: RotationMatrixTensor) -> QuaternionTensor:
    """
    Convert rotations given as rotation matrices to quaternions.
    Args:
        matrix: Rotation matrices as tensor of shape ``(..., 3, 3)``.
    Returns:
        quaternions with real part first, as tensor of shape ``(..., 4)``.
    """
    if matrix.size(-1) != 3 or matrix.size(-2) != 3:
        raise ValueError(f"Invalid rotation matrix shape {matrix.shape}.")

    batch_dim = matrix.shape[:-2]
    m00, m01, m02, m10, m11, m12, m20, m21, m22 = torch.unbind(
        matrix.reshape(batch_dim + (9,)), dim=-1
    )

    q_abs = _sqrt_positive_part(
        torch.stack(
            [
                1.0 + m00 + m11 + m22,
                1.0 + m00 - m11 - m22,
                1.0 - m00 + m11 - m22,
                1.0 - m00 - m11 + m22,
            ],
            dim=-1,
        )
    )

    # we produce the desired quaternion multiplied by each of r, i, j, k
    quat_by_rijk = torch.stack(
        [
            # pyre-fixme[58]: `**` is not supported for operand types `Tensor` and
            #  `int`.
            torch.stack(
                [q_abs[..., 0] ** 2, m21 - m12, m02 - m20, m10 - m01], dim=-1
            ),
            # pyre-fixme[58]: `**` is not supported for operand types `Tensor` and
            #  `int`.
            torch.stack(
                [m21 - m12, q_abs[..., 1] ** 2, m10 + m01, m02 + m20], dim=-1
            ),
            # pyre-fixme[58]: `**` is not supported for operand types `Tensor` and
            #  `int`.
            torch.stack(
                [m02 - m20, m10 + m01, q_abs[..., 2] ** 2, m12 + m21], dim=-1
            ),
            # pyre-fixme[58]: `**` is not supported for operand types `Tensor` and
            #  `int`.
            torch.stack(
                [m10 - m01, m20 + m02, m21 + m12, q_abs[..., 3] ** 2], dim=-1
            ),
        ],
        dim=-2,
    )

    # We floor here at 0.1 but the exact level is not important; if q_abs is small,
    # the candidate won't be picked.
    flr = torch.tensor(0.1).to(dtype=q_abs.dtype, device=q_abs.device)
    quat_candidates = quat_by_rijk / (2.0 * q_abs[..., None].max(flr))

    # if not for numerical problems, quat_candidates[i] should be same (up to a sign),
    # forall i; we pick the best-conditioned one (with the largest denominator)

    return quat_candidates[
        F.one_hot(q_abs.argmax(dim=-1), num_classes=4) > 0.5, :
    ].reshape(batch_dim + (4,))


def quaternion_to_matrix(
    quaternions: QuaternionTensor,
) -> RotationMatrixTensor:
    """
    Convert rotations given as quaternions to rotation matrices.

    Args:
        quaternions: quaternions with real part first,
            as tensor of shape ``(..., 4)``.

    Returns:
        Rotation matrices as tensor of shape ``(..., 3, 3)``.
    """
    r, i, j, k = torch.unbind(quaternions, -1)
    # pyre-fixme[58]: `/` is not supported for operand types `float` and `Tensor`.
    two_s = 2.0 / (quaternions * quaternions).sum(-1)

    o = torch.stack(
        (
            1 - two_s * (j * j + k * k),
            two_s * (i * j - k * r),
            two_s * (i * k + j * r),
            two_s * (i * j + k * r),
            1 - two_s * (i * i + k * k),
            two_s * (j * k - i * r),
            two_s * (i * k - j * r),
            two_s * (j * k + i * r),
            1 - two_s * (i * i + j * j),
        ),
        -1,
    )
    return o.reshape(quaternions.shape[:-1] + (3, 3))


def idealize_backbone(
    x: Union[AtomTensor, BackboneTensor],
    lr: float = 1e-3,
    n_iter: int = 100,
    inplace: bool = False,
) -> BackboneTensor:
    """Idealizes a protein backbone to more closely resemble idealised geometry.

    Adaptation of an implementation by Sergey Ovchinnikov:
    https://github.com/sokrypton/tf_proteins/blob/master/coord_to_dihedrals_tools.ipynb

    :param x: Tensor representing the backbone (can include sidechain atoms but
        these are not used).
    :type x: Union[AtomTensor, BackboneTensor]
    :param lr: Learning rate to use with the optimiser (Adam), defaults to
        ``1e-3``.
    :type lr: float, optional
    :param n_iter: Number of optimisation steps to make, defaults to ``100``.
    :type n_iter: int, optional
    :return: BackboneTensor with idealised geometry.
    :rtype: BackboneTensor
    """
    if not inplace:
        x = copy.deepcopy(x)
    x_constant = x
    x.requires_grad = True

    def mse(x, y):
        return torch.mean((x - y) ** 2)

    def ideal_loss(x, x_constant) -> torch.Tensor:
        bl = get_backbone_bond_lengths(x)
        ba = get_backbone_bond_angles(x)

        loss_ideal = torch.sum(
            torch.norm(
                torch.tensor(IDEAL_BB_BOND_LENGTHS, device=bl.device) - bl,
                dim=-1,
            )
        ) + torch.sum(
            torch.norm(
                torch.tensor(IDEAL_BB_BOND_ANGLES, device=ba.device) - ba
            ),
            dim=-1,
        )
        loss_ms = mse(x[:, :4, :], x_constant[:, :4, :])

        return loss_ideal + loss_ms

    opt = torch.optim.Adam([x], lr=lr)

    for _ in range(n_iter):
        loss = ideal_loss(x, x_constant)
        opt.zero_grad()
        loss.backward()
        opt.step()
    return x


def apply_structural_noise(
    x: Union[AtomTensor, CoordTensor],
    magnitude: float = 0.1,
    gaussian: bool = True,
    return_transformed: bool = True,
) -> Union[AtomTensor, CoordTensor]:
    """
    Generates random noise and adds it to the input tensor.

    :param x: Input tensor
    :param magnitude: Magnitude of the noise
    :param gaussian: If ``True``, noise is sampled from a Gaussian distribution.
        If ``False``, noise is sampled from a uniform distribution. Default is
        ``True`` (gaussian).
    :param return_transformed: If ``True``, returns the transformed (noised)
        tensor. If ``False``, returns the noise tensor. Default is ``True``.
    :return: Noised tensor or noise tensor
    """
    if gaussian:
        noise = torch.randn_like(x, device=x.device) * magnitude
    else:
        noise = (torch.rand_like(x, device=x.device) - 0.5) * 2 * magnitude
    return x + noise if return_transformed else noise



"""Functions to add embeddings from pre-trained language models protein
structure graphs."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import os
from functools import lru_cache, partial
from pathlib import Path

import networkx as nx
import numpy as np
from loguru import logger as log

from graphein.protein.features.sequence.utils import (
    compute_feature_over_chains,
    subset_by_node_feature_value,
)
from graphein.utils.dependencies import import_message

try:
    import torch
except ImportError:
    message = import_message(
        submodule="graphein.protein.features.sequence.embeddings",
        package="torch",
        pip_install=True,
        conda_channel="pytorch",
    )
    log.debug(message)

try:
    import biovec
except ImportError:
    message = import_message(
        submodule="graphein.protein.features.sequence.embeddings",
        package="biovec",
        pip_install=True,
        extras=True,
    )
    log.debug(message)


@lru_cache()
def _load_esm_model(model_name: str = "esm1b_t33_650M_UR50S"):
    """
    Loads pre-trained FAIR ESM model from torch hub.

        *Biological Structure and Function Emerge from Scaling Unsupervised*
        *Learning to 250 Million Protein Sequences* (2019)
        Rives, Alexander and Meier, Joshua and Sercu, Tom and Goyal, Siddharth
        and Lin, Zeming and Liu, Jason and Guo, Demi and Ott, Myle and Zitnick,
        C. Lawrence and Ma, Jerry and Fergus, Rob


        *Transformer protein language models are unsupervised structure learners*
        2020 Rao, Roshan M and Meier, Joshua and Sercu, Tom and Ovchinnikov,
        Sergey and Rives, Alexander

    Pre-trained models:
    Full Name layers params Dataset Embedding Dim Model URL
    ========= ====== ====== ======= ============= =========
    ESM-1b   esm1b_t33_650M_UR50S 33 650M UR50/S 1280 https://dl.fbaipublicfiles.com/fair-esm/models/esm1b_t33_650M_UR50S.pt
    ESM1-main esm1_t34_670M_UR50S34 670M UR50/S 1280 https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t34_670M_UR50S.pt
    esm1_t34_670M_UR50D 34 670M UR50/D 1280 https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t34_670M_UR50D.pt
    esm1_t34_670M_UR100 34 670M UR100 1280 https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t34_670M_UR100.pt
    esm1_t12_85M_UR50S 12 85M UR50/S 768 https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t12_85M_UR50S.pt
    esm1_t6_43M_UR50S 6 43M UR50/S 768 https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t6_43M_UR50S.pt

    :param model_name: Name of pre-trained model to load
    :type model_name: str
    :return: loaded pre-trained model
    """

    return torch.hub.load("facebookresearch/esm", model_name)


def compute_esm_embedding(
    sequence: str,
    representation: str,
    model_name: str = "esm1b_t33_650M_UR50S",
    output_layer: int = 33,
) -> np.ndarray:
    """
    Computes sequence embedding using Pre-trained ESM model from FAIR

        *Biological Structure and Function Emerge from Scaling Unsupervised*
        *Learning to 250 Million Protein Sequences* (2019)
        Rives, Alexander and Meier, Joshua and Sercu, Tom and Goyal, Siddharth
        and Lin, Zeming and Liu, Jason and Guo, Demi and Ott, Myle and Zitnick,
        C. Lawrence and Ma, Jerry and Fergus, Rob

        *Transformer protein language models are unsupervised structure learners*
        2020 Rao, Roshan M and Meier, Joshua and Sercu, Tom and Ovchinnikov,
            Sergey and Rives, Alexander

    Pre-trained models:

    Full Name layers params Dataset Embedding Dim Model URL
    ========= ====== ====== ======= ============= =========
    ESM-1b esm1b_t33_650M_UR50S 33 650M UR50/S 1280 https://dl.fbaipublicfiles.com/fair-esm/models/esm1b_t33_650M_UR50S.pt
    ESM1-main esm1_t34_670M_UR50S 34 670M UR50/S 1280 https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t34_670M_UR50S.pt
    esm1_t34_670M_UR50D 34 670M UR50/D 1280 https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t34_670M_UR50D.pt
    esm1_t34_670M_UR100 34 670M UR100 1280 https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t34_670M_UR100.pt
    esm1_t12_85M_UR50S 12 85M UR50/S 768 https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t12_85M_UR50S.pt
    esm1_t6_43M_UR50S 6 43M UR50/S 768 https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t6_43M_UR50S.pt

    :param sequence: Protein sequence to embed (str)
    :type sequence: str
    :param representation: Type of embedding to extract. ``"residue"`` or
        ``"sequence"``. Sequence-level embeddings are averaged residue
        embeddings
    :type representation: str
    :param model_name: Name of pre-trained model to use
    :type model_name: str
    :param output_layer: integer indicating which layer the output should be
        taken from.
    :type output_layer: int
    :return: embedding (``np.ndarray``)
    :rtype: np.ndarray
    """
    model, alphabet = _load_esm_model(model_name)
    batch_converter = alphabet.get_batch_converter()

    data = [
        ("protein1", sequence),
    ]
    batch_labels, batch_strs, batch_tokens = batch_converter(data)

    # Extract per-residue representations (on CPU)
    with torch.no_grad():
        results = model(
            batch_tokens, repr_layers=[output_layer], return_contacts=True
        )
    token_representations = results["representations"][output_layer]

    if representation == "residue":
        return token_representations.numpy()

    # Generate per-sequence representations via averaging
    # NOTE: token 0 is always a beginning-of-sequence token, so the first
    # residue is token 1.
    elif representation == "sequence":
        sequence_representations = []
        for i, (_, seq) in enumerate(data):
            sequence_representations.append(
                token_representations[i, 1 : len(seq) + 1].mean(0)
            )
        return sequence_representations[0].numpy()


def esm_residue_embedding(
    G: nx.Graph,
    model_name: str = "esm1b_t33_650M_UR50S",
    output_layer: int = 33,
) -> nx.Graph:
    """
    Computes ESM residue embeddings from a protein sequence and adds the to the
    graph.

        *Biological Structure and Function Emerge from Scaling Unsupervised*
        *Learning to 250 Million Protein Sequences* (2019)
        Rives, Alexander and Meier, Joshua and Sercu, Tom and Goyal, Siddharth
        and Lin, Zeming and Liu, Jason and Guo,
        Demi and Ott, Myle and Zitnick, C. Lawrence and Ma, Jerry and Fergus,
        Rob


        *Transformer protein language models are unsupervised structure learners*
        (2020) Rao, Roshan M and Meier, Joshua and Sercu, Tom and Ovchinnikov,
        Sergey and Rives, Alexander

    **Pre-trained models**

    =========                     ====== ====== ======= ============= =========
    Full Name                     layers params Dataset Embedding Dim Model URL
    =========                     ====== ====== ======= ============= =========
    ESM-1b esm1b_t33_650M_UR50S   33    650M   UR50/S     1280        https://dl.fbaipublicfiles.com/fair-esm/models/esm1b_t33_650M_UR50S.pt
    ESM1-main esm1_t34_670M_UR50S 34    670M   UR50/S     1280        https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t34_670M_UR50S.pt
    esm1_t34_670M_UR50D           34    670M   UR50/D     1280        https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t34_670M_UR50D.pt
    esm1_t34_670M_UR100           34    670M   UR100      1280        https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t34_670M_UR100.pt
    esm1_t12_85M_UR50S            12    85M    UR50/S     768         https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t12_85M_UR50S.pt
    esm1_t6_43M_UR50S             6     43M    UR50/S     768         https://dl.fbaipublicfiles.com/fair-esm/models/esm1_t6_43M_UR50S.pt
    =========                     ====== ====== ======= ============= =========

    :param G: ``nx.Graph`` to add esm embedding to.
    :type G: nx.Graph
    :param model_name: Name of pre-trained model to use.
    :type model_name: str
    :param output_layer: index of output layer in pre-trained model.
    :type output_layer: int
    :return: ``nx.Graph`` with esm embedding feature added to nodes.
    :rtype: nx.Graph
    """

    for chain in G.graph["chain_ids"]:
        embedding = compute_esm_embedding(
            G.graph[f"sequence_{chain}"],
            representation="residue",
            model_name=model_name,
            output_layer=output_layer,
        )
        # remove start and end tokens from per-token residue embeddings
        embedding = embedding[0, 1:-1]
        subgraph = subset_by_node_feature_value(G, "chain_id", chain)

        for i, (n, d) in enumerate(subgraph.nodes(data=True)):
            G.nodes[n]["esm_embedding"] = embedding[i]

    return G


def esm_sequence_embedding(G: nx.Graph) -> nx.Graph:
    """
    Computes ESM sequence embedding feature over chains in a graph.

    :param G: nx.Graph protein structure graph.
    :type G: nx.Graph
    :return: nx.Graph protein structure graph with esm embedding features added
        eg. ``G.graph["esm_embedding_A"]`` for chain A.
    :rtype: nx.Graph
    """
    func = partial(compute_esm_embedding, representation="sequence")
    G = compute_feature_over_chains(G, func, feature_name="esm_embedding")

    return G


@lru_cache()
def _load_biovec_model():
    """Loads pretrained ProtVec Model.

    **Source**

        ProtVec: A Continuous Distributed Representation of Biological Sequences

    Paper: http://arxiv.org/pdf/1503.05140v1.pdf
    """

    return biovec.models.load_protvec(
        os.fspath(
            Path(__file__).parent.parent
            / "pretrained_models"
            / "swissprot-reviewed-protvec.model"
        )
    )


def biovec_sequence_embedding(G: nx.Graph) -> nx.Graph:
    """
    Adds BioVec sequence embedding feature to the graph. Computed over chains.

    **Source**
        ProtVec: A Continuous Distributed Representation of Biological Sequences

    Paper: http://arxiv.org/pdf/1503.05140v1.pdf

    :param G: nx.Graph protein structure graph.
    :type G: nx.Graph
    :return: nx.Graph protein structure graph with biovec embedding added. e.g.
        ``G.graph["biovec_embedding_A"]`` for chain ``A``.
    :rtype: nx.Graph
    """
    pv = _load_biovec_model()
    func = pv.to_vecs
    G = compute_feature_over_chains(G, func, feature_name="biovec_embedding")
    return G



"""Functions for adding metadata to PPI Graphs from STRING and BIOGRID."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Ramon Vinas
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from typing import Dict, Union

import networkx as nx
import pandas as pd
from loguru import logger as log

from graphein.ppi.parse_biogrid import BIOGRID_df
from graphein.ppi.parse_stringdb import STRING_df


def add_string_metadata(
    G: nx.Graph, kwargs: Dict[str, Union[str, int]]
) -> nx.Graph:
    """
    Adds interaction dataframe from STRING to graph.

    :param G: PPI Graph to add metadata to.
    :type G: nx.Graph
    :param kwargs:  Additional parameters for STRING API call.
    :type kwargs: Dict[str, Union[str, int]]
    :return: PPIGraph with added STRING ``interaction_df`` as metadata.
    :rtype: nx.Graph
    """
    G.graph["string_df"] = STRING_df(
        G.graph["protein_list"], G.graph["ncbi_taxon_id"], kwargs
    )
    log.debug("Added STRING interaction dataframe as graph metadata")
    return G


def add_biogrid_metadata(
    G: nx.Graph, kwargs: Dict[str, Union[str, int]]
) -> nx.Graph:
    """
    Adds interaction dataframe from BIOGRID to graph.

    :param G: PPI Graph to add metadata to
    :type G: nx.Graph
    :param kwargs:  Additional parameters for BIOGRID API call
    :type kwargs: Dict[str, Union[str, int]]
    :return: PPIGraph with added BIOGRID interaction_df as metadata
    :rtype: nx.Graph
    """
    G.graph["string_df"] = BIOGRID_df(
        G.graph["protein_list"], G.graph["ncbi_taxon_id"], kwargs
    )
    log.debug("Added BIOGRID interaction dataframe as graph metadata")
    return G


def add_string_biogrid_metadata(
    G: nx.Graph, kwargs: Dict[str, Union[str, int]]
) -> nx.Graph:
    """
    Adds interaction dataframe from STRING and BIOGRID to graph.

    :param G: PPIGraph to add metadata to.
    :type G: nx.Graph
    :param kwargs:  Additional parameters for STRING and BIOGRID API calls.
    :type kwargs: Dict[str, Union[str, int]]
    :return: PPIGraph with added STRING and BIOGRID interaction_df as metadata.
    :rtype: nx.Graph
    """
    G.graph["string_df"] = STRING_df(
        G.graph["protein_list"], G.graph["ncbi_taxon_id"], kwargs
    )
    G.graph["biogrid_df"] = BIOGRID_df(
        G.graph["protein_list"], G.graph["ncbi_taxon_id"], kwargs
    )
    G.graph["combined_interaction_df"] = pd.concat(
        [G.graph["string_df"], G.graph["biogrid_df"]]
    )
    log.debug(
        "Added combined STRING and BIOGRID interaction dataframe as graph metadata"
    )
    return G



# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: BSD 3 clause
# Code Repository: https://github.com/a-r-j/graphein
from loguru import logger
from rich.logging import RichHandler

from graphein.utils.utils import *

# from .protein import *
# from .rna import *
from .testing import *

__author__ = "Arian Jamasb <arian@jamasb.io>"
__version__ = "1.7.8"


logger.configure(
    handlers=[
        {"sink": RichHandler(rich_tracebacks=True), "format": "{message}"}
    ]
)

logger.disable("graphein")


def verbose(enabled: bool = False):
    """Enable/Disable logging.

    :param enabled: Whether or not to enable logging, defaults to ``False``.
    :type enabled: bool, optional
    """
    if not enabled:
        logger.disable("graphein")
    else:
        logger.enable("graphein")



"""Functions for featurising Small Molecule Graphs."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

from typing import Dict, List, Optional, Union

import networkx as nx
import numpy as np
import pandas as pd

from graphein.molecule.utils import (
    count_fragments,
    get_center,
    get_max_ring_size,
    get_mol_weight,
    get_morgan_fp,
    get_morgan_fp_np,
    get_qed_score,
    get_shape_moments,
)
from graphein.utils.dependencies import import_message

try:
    from rdkit import Chem
    from rdkit.Chem import Descriptors
except ImportError:
    import_message(
        "graphe in.molecule.features.graph.molecule",
        "rdkit",
        "rdkit",
        True,
        extras=True,
    )


def mol_descriptors(
    g: nx.Graph,
    descriptor_list: Optional[List[str]] = None,
    return_array: bool = False,
    return_series: bool = False,
) -> Union[np.ndarray, pd.Series, Dict[str, Union[float, int]]]:
    """Adds global molecular descriptors to the graph.

    :param g: The graph to add the descriptors to.
    :type g: nx.Graph
    :param descriptor_list: The list of descriptors to add. If ``None``, all descriptors are added.
    :type descriptor_list: Optional[List[str]]
    :param return_array: If ``True``, the descriptors are returned as a ``np.ndarray``.
    :type return_array: bool
    :param return_series: If ``True``, the descriptors are returned as a ``pd.Series``.
    :return: The descriptors as a dictionary (default) ``np.ndarray`` or ``pd.Series``.
    :rtype: Union[np.ndarray, pd.Series,  Dict[str, Union[float, int]]]
    """

    mol = g.graph["rdmol"]
    # Retrieve list of possible descriptors
    descriptors = {d[0]: d[1] for d in Descriptors.descList}

    # Subset descriptors to those provided
    if descriptor_list is not None:
        descriptors = {
            k: v for k, v in descriptors.items() if k in descriptor_list
        }

    # Compute descriptors
    desc = {d: descriptors[d](mol) for d in descriptors}

    # Process Outformat
    if return_array:
        desc = np.array(desc.values())
        g.graph["descriptors"] = desc
    elif return_series:
        desc = pd.Series(desc)
        g.graph["descriptors"] = desc
    else:
        g.graph.update(desc)

    return desc


def add_center_of_mol(
    g: nx.Graph, weights: Optional[np.ndarray] = None
) -> nx.Graph:
    """Compute the centroid of the conformation.

    Hydrogens are ignored and no attention is paid to the difference in sizes
    of the heavy atoms; however, an optional vector of weights can be passed.

    :param g: Molecular Graph or RDkit Mol to compute the center of.
    :type g: nx.Graph
    :return: Graph with the centroid of the molecule added as a graph feature.
    :rtype: np.ndarray
    """
    g.graph["center"] = get_center(g, weights=weights)
    return g


def add_shape_moments(g: nx.Graph):
    """Add principal moments of inertia as defined in https://pubs.acs.org/doi/10.1021/ci025599w

    :param mol: Molecular Graph or RDkit Mol to compute the moments of intertia of.
    :type mol: Union[nx.Graph, Chem.Mol]
    :return: First 2 moments as a tuple.
    :rtype: Tuple[float, float]
    """
    g.graph["shape_moments"] = get_shape_moments(g)
    return g


def add_max_ring_size(g: nx.Graph) -> nx.Graph:
    g.graph["max_ring_size"] = get_max_ring_size(g)
    return g


def add_morgan_fingerprint(
    g: nx.Graph, radius: int = 2, n_bits: int = 2048
) -> nx.Graph:
    g.graph["morgan_fingerprint"] = get_morgan_fp(
        g, radius=radius, n_bits=n_bits
    )
    return g


def add_morgan_fingerprint_np(
    g: nx.Graph, radius: int = 2, n_bits: int = 2048
) -> nx.Graph:
    g.graph["morgan_fingerprint_np"] = get_morgan_fp_np(
        g, radius=radius, n_bits=n_bits
    )
    return g


def add_fragment_counts(g: nx.Graph) -> nx.Graph:
    g.graph["fragment_counts"] = count_fragments(g)
    return g


def add_mol_weight(g: nx.Graph) -> nx.Graph:
    g.graph["mol_weight"] = get_mol_weight(g)
    return g


def add_qed_score(g: nx.Graph) -> nx.Graph:
    g.graph["qed"] = get_qed_score(g)
    return g



"""Yaml parser for config objects"""

from functools import partial

from deepdiff.operator import BaseOperator


def partial_functions_equal(func1: partial, func2: partial) -> bool:
    """
    Determine whether two partial functions are equal.

    :param func1: Partial function to check
    :type func1: partial
    :param func2: Partial function to check
    :type func2: partial
    :return: Whether the two functions are equal
    :rtype: bool
    """
    return (
        all(
            getattr(func1, attr) == getattr(func2, attr)
            for attr in ["func", "args", "keywords"]
        )
        if (isinstance(func1, partial) and isinstance(func2, partial))
        else False
    )


class PartialMatchOperator(BaseOperator):
    """Custom operator for deepdiff comparison. This operator compares whether the two partials are equal."""

    def give_up_diffing(self, level, diff_instance):
        return partial_functions_equal(level.t1, level.t2)


class PathMatchOperator(BaseOperator):
    """Custom operator for deepdiff comparison. This operator compares whether the two pathlib Paths are equal."""

    def give_up_diffing(self, level, diff_instance):
        return level.t1 == level.t2



"""Tests for graphein.protein.tensor.representation."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import pytest

from graphein.protein.tensor.geometry import kabsch
from graphein.protein.tensor.reconstruction import dist_mat_to_coords
from graphein.protein.tensor.representation import (
    get_c_alpha,
    get_full_atom_coords,
)

try:
    import torch

    TORCH_AVAIL = True
except ImportError:
    TORCH_AVAIL = False


@pytest.mark.skipif(not TORCH_AVAIL, reason="PyTorch not available")
def test_get_c_alpha():
    """Test get_c_alpha"""
    x = torch.rand((2, 37, 3))
    assert get_c_alpha(x).shape == (2, 3)
    assert torch.allclose(get_c_alpha(x), x[:, 1, :])

    x = torch.rand((2, 3))
    assert get_c_alpha(x).shape == (2, 3)
    assert torch.allclose(get_c_alpha(x), x)


@pytest.mark.skipif(not TORCH_AVAIL, reason="PyTorch not available")
def test_get_full_atom_coords():
    # Test case 1
    atom_tensor = torch.Tensor(
        [[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [10, 11, 12]]]
    )
    coords, residue_index, atom_type = get_full_atom_coords(atom_tensor)
    assert coords.tolist() == [[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]]
    assert residue_index.tolist() == [0, 0, 1, 1]
    assert atom_type.tolist() == [0, 1, 0, 1]

    # Test case 2
    atom_tensor = torch.Tensor(
        [[[1, 2, 3], [4, 5, 6]], [[1e-5, 1e-5, 1e-5], [1e-5, 1e-5, 1e-5]]]
    )
    coords, residue_index, atom_type = get_full_atom_coords(atom_tensor)
    assert coords.tolist() == [[1, 2, 3], [4, 5, 6]]
    assert residue_index.tolist() == [0, 0]
    assert atom_type.tolist() == [0, 1]

    # Test case 3
    atom_tensor = torch.Tensor(
        [[[1, 2, 3], [4, 5, 6]], [[1e-5, 1e-5, 1e-5], [10, 11, 12]]]
    )
    coords, residue_index, atom_type = get_full_atom_coords(atom_tensor)
    assert coords.tolist() == [[1, 2, 3], [4, 5, 6], [10, 11, 12]]
    assert residue_index.tolist() == [0, 0, 1]
    assert atom_type.tolist() == [0, 1, 1]

    # Test case 4
    atom_tensor = torch.Tensor(
        [[[1, 2, 3], [4, 5, 6]], [[1e-6, 1e-6, 1e-6], [1e-6, 1e-6, 1e-6]]]
    )
    coords, residue_index, atom_type = get_full_atom_coords(
        atom_tensor, fill_value=1e-6
    )
    assert coords.tolist() == [[1, 2, 3], [4, 5, 6]]
    assert residue_index.tolist() == [0, 0]
    assert atom_type.tolist() == [0, 1]



"""Base Config object for use with Molecule Graph Construction."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Any, Callable, List, Optional, Union

from deepdiff import DeepDiff
from pydantic import BaseModel

from graphein.molecule.edges.atomic import add_atom_bonds
from graphein.molecule.edges.distance import (
    add_distance_threshold,
    add_fully_connected_edges,
    add_k_nn_edges,
)
from graphein.molecule.features.nodes.atom_type import atom_type_one_hot
from graphein.utils.config import PartialMatchOperator, PathMatchOperator

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


GraphAtoms = Literal[
    "C",
    "H",
    "O",
    "N",
    "F",
    "P",
    "S",
    "Cl",
    "Br",
    "I",
    "B",
]
"""Allowable atom types for nodes in the graph."""


class MoleculeGraphConfig(BaseModel):
    """
    Config Object for Molecule Structure Graph Construction.

    :param verbose: Specifies verbosity of graph creation process.
    :type verbose: bool
    :param add_hs: Specifies whether hydrogens should be added to the graph.
    :type add_hs: bool
    :param generate_conformer: Specifies whether or not to generate a conformer for inputs that do not provide coordinates (e.g. SMILES).
        Uses :meth:`graphein.molecule.graphs.generate_3d` to generate conformers.
        Default: ``False``.
    :param edge_construction_functions: List of functions that take an ``nx.Graph`` and return an ``nx.Graph`` with desired
        edges added. Prepared edge constructions can be found in :ref:`graphein.protein.edges`
    :type edge_construction_functions: List[Callable]
    :param node_metadata_functions: List of functions that take an ``nx.Graph``
    :type node_metadata_functions: List[Callable], optional
    :param edge_metadata_functions: List of functions that take an
    :type edge_metadata_functions: List[Callable], optional
    :param graph_metadata_functions: List of functions that take an ``nx.Graph`` and return an ``nx.Graph`` with added
        graph-level features and metadata.
    :type graph_metadata_functions: List[Callable], optional
    """

    verbose: bool = False
    add_hs: bool = False
    generate_conformer: bool = False
    # Graph construction functions
    edge_construction_functions: List[Union[Callable, str]] = [
        add_fully_connected_edges,
        add_k_nn_edges,
        add_distance_threshold,
        add_atom_bonds,
    ]
    node_metadata_functions: Optional[List[Union[Callable, str]]] = [
        atom_type_one_hot
    ]
    edge_metadata_functions: Optional[List[Union[Callable, str]]] = None
    graph_metadata_functions: Optional[List[Callable]] = None

    def __eq__(self, other: Any) -> bool:
        """Overwrites the BaseModel __eq__ function in order to check more specific cases (like partial functions)."""
        if isinstance(other, MoleculeGraphConfig):
            return (
                DeepDiff(
                    self,
                    other,
                    custom_operators=[
                        PartialMatchOperator(types=[partial]),
                        PathMatchOperator(types=[Path]),
                    ],
                )
                == {}
            )
        return self.dict() == other



"""Visualisation utilities for RNA Secondary Structure Graphs."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from collections import defaultdict
from itertools import chain
from typing import Dict, List

import networkx as nx
from loguru import logger as log

from graphein.protein.visualisation import (
    plotly_protein_structure_graph as plotly_rna_structure_graph,
)
from graphein.rna.graphs import RNA_BASE_COLORS


def plot_rna_graph(
    g: nx.Graph,
    layout: nx.layout = nx.layout.circular_layout,
    label_base_type: bool = True,
    label_base_position: bool = False,
    label_dotbracket_symbol: bool = False,
    **kwargs,
):
    """Plots a RNA Secondary Structure Graph. Colours edges by kind.

    :param g: NetworkX graph of RNA secondary structure graph.
    :type g: nx.Graph
    :param layout: Layout algorithm to use. Default is circular_layout.
    :type layout: nx.layout
    :param label_base_type: Whether to label the base type of each base.
    :type label_base_type: bool
    :param label_base_position: Whether to label the base position of each base.
    :type label_base_position: bool
    :param label_dotbracket_symbol: Whether to label the dotbracket symbol of each base.
    """
    edge_colors = nx.get_edge_attributes(g, "color").values()
    node_colors = nx.get_node_attributes(g, "color").values()
    if len(node_colors) == 0:
        node_colors = None
    if len(edge_colors) == 0:
        edge_colors = None

    # Construct node labelling scheme
    node_label_dicts: List[Dict[int, str]] = []
    if label_base_type:
        try:
            node_label_dicts.append(
                {n: d["nucleotide"] for n, d in g.nodes(data=True)}
            )
        except KeyError:
            log.warning(
                "No sequence data found in graph. Skipping base type labelling."
            )

    if label_base_position:
        node_label_dicts.append({n: str(n) for n in g.nodes()})
    elif label_dotbracket_symbol:
        node_label_dicts.append(
            {n: d["dotbracket_symbol"] for n, d in g.nodes(data=True)}
        )

    node_labels = defaultdict(str)
    for key, value in chain.from_iterable(map(dict.items, node_label_dicts)):
        node_labels[key] += value

    nx.draw(
        g,
        node_color=node_colors,
        edge_color=edge_colors,
        with_labels=True,
        labels=node_labels,
        pos=layout(g),
        **kwargs,
    )



"""Functions to compute edges for an RNA secondary structure graph."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Emmanuele Rossi, Eric Ma
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import networkx as nx
from loguru import logger as log

from graphein.rna.constants import (
    CANONICAL_BASE_PAIRINGS,
    PSEUDOKNOT_CLOSING_SYMBOLS,
    PSEUDOKNOT_OPENING_SYMBOLS,
    SIMPLE_DOTBRACKET_NOTATION,
    WOBBLE_BASE_PAIRINGS,
)

BASEPAIRING_EDGE_FUNCS = [
    "add_phosphodiester_bonds",
    "add_base_pairing_interactions",
    "add_pseudoknots",
    "add_base_pairing_interactions",
]


def check_base_pairing_type(base_1: str, base_2: str) -> str:
    """
    Checks type and validity of base pairing interactions.

    :param base_1: str RNA Base letter for base 1.
    :type base_1: str
    :param base_2: str RNA base letter for base 2.
    :type base_2: str
    :return: string referencing the type of base pairing ``"canonical"``,
        ``"wobble"`` or ``"invalid"``.
    :rtype: str
    """
    try:
        if base_2 in CANONICAL_BASE_PAIRINGS[base_1]:
            return "canonical"
        elif base_2 in WOBBLE_BASE_PAIRINGS[base_1]:
            return "wobble"
    except KeyError:
        return "invalid"


def add_phosphodiester_bonds(G: nx.Graph) -> nx.Graph:
    """
    Adds phosphodiester bonds between adjacent nucleotides to an RNA secondary
    structure graph.

    :param G: RNA Graph to add edges to.
    :type G: nx.Graph
    :return: RNA graph with ``phosphodiester_bond`` edges added.
    :rtype: nx.Graph
    """
    # Iterate over dotbracket to build connectivity
    bases = []
    for i, c in enumerate(G.graph["dotbracket"]):
        # Add adjacent edges (phosphodiester_bonds)
        if i > 0:
            G.add_edge(i, i - 1, attr="phosphodiester_bond", color="b")
    log.debug("Added phosphodiester bonds as edges")
    return G


def add_base_pairing_interactions(G: nx.Graph) -> nx.Graph:
    """
    Adds base pairing interactions between nucleotides to an RNA secondary
    structure graph.

    :param G: RNA Graph to add edges to.
    :type G: nx.Graph
    :raises ValueError: if ``dotbracket`` contains an unsupported character.
    :return: RNA graph with ``base_pairing`` edges added.
    :rtype: nx.Graph
    """
    # Check sequence is used
    check_base_pairing = "sequence" in G.graph.keys()
    # Iterate over dotbracket to build connectivity
    bases = []
    for i, c in enumerate(G.graph["dotbracket"]):
        # Add base_pairing interactions
        if c == "(":
            bases.append(i)
        elif c == ")":
            neighbor = bases.pop()

            if check_base_pairing:
                pairing_type = check_base_pairing_type(
                    G.nodes[i]["nucleotide"], G.nodes[neighbor]["nucleotide"]
                )
            else:
                pairing_type = "unknown"

            G.add_edge(
                i,
                neighbor,
                attr="base_pairing",
                pairing_type=pairing_type,
                color="r",
            )
        elif c in [".", "[", "]", "{", "}", "<", ">"]:
            continue
        else:
            raise ValueError("Input is not in dot-bracket notation!")
        log.debug("Added base_pairing interactions as edges")
    return G


def add_pseudoknots(G: nx.Graph) -> nx.Graph:
    """
    Adds pseudoknots nucleotides to an RNA secondary structure graph.

    :param G: RNA Graph to add edges to.
    :type G: nx.Graph
    :return: RNA graph with pseudoknot edges added.
    :rtype: nx.Graph
    """
    # Check sequence is used
    check_base_pairing = "sequence" in G.graph.keys()
    # Iterate over dotbracket to build connectivity
    knot_bases_1 = []  # for [[[]]] knots
    knot_bases_2 = []  # for {{{}}} knots
    knot_bases_3 = []  # for <<<>>> knots

    for i, c in enumerate(G.graph["dotbracket"]):
        if c in PSEUDOKNOT_OPENING_SYMBOLS:
            if c == "<":
                knot_bases_3.append(i)
            elif c == "[":
                knot_bases_1.append(i)
            elif c == "{":
                knot_bases_2.append(i)
        elif c in PSEUDOKNOT_CLOSING_SYMBOLS:
            if c == ">":
                neighbor = knot_bases_3.pop()
            elif c == "]":
                neighbor = knot_bases_1.pop()
            elif c == "}":
                neighbor = knot_bases_2.pop()
            if check_base_pairing:
                pairing_type = check_base_pairing_type(
                    G.nodes[i]["nucleotide"], G.nodes[neighbor]["nucleotide"]
                )
            else:
                pairing_type = "unknown"
            G.add_edge(
                i,
                neighbor,
                attr="pseudoknot",
                pairing_type=pairing_type,
                color="g",
            )
        elif c in SIMPLE_DOTBRACKET_NOTATION:
            continue
        else:
            raise ValueError("Input is not in dot-bracket notation!")
        log.debug("Added pseudoknot interactions as edges")
    return G


def add_all_dotbracket_edges(G: nx.Graph) -> nx.Graph:
    """
    Adds phosphodiester bonds between adjacent nucleotides and base_pairing
    interactions to an RNA secondary structure graph.

    :param G: RNA Graph to add edges to.
    :type G: nx.Graph
    :return: RNA graph with ``phosphodiester_bond`` and ``base_pairing`` edges
        added.
    :rtype: nx.Graph
    """
    # Iterate over dotbracket to build connectivity
    G = add_phosphodiester_bonds(G)
    G = add_base_pairing_interactions(G)
    G = add_pseudoknots(G)
    return G



"""Utilities for sanity checking protein tensors."""

from functools import partial
from typing import List, Tuple, Union

from loguru import logger as log

from graphein.utils.dependencies import import_message

from ..resi_atoms import (
    ATOM_NUMBERING_MODIFIED,
    STANDARD_AMINO_ACID_MAPPING_1_TO_3,
)
from .sequence import get_atom_indices
from .types import AtomTensor, BackboneTensor, CoordTensor, ResidueTensor

try:
    import torch
except ImportError:
    message = import_message(
        "graphein.protein.tensor.testing",
        "torch",
        conda_channel="pytorch",
        pip_install=True,
    )
    log.warning(message)


def has_nan(x: torch.Tensor) -> bool:
    """Checks a tensor for ``NaN`` values.

    :param x: Tensor to check for ``NaN`` values
    :type x: torch.Tensor
    :return: Boolean indicating whether the tensor contains NaN values.
    :rtype: bool
    """
    return x.isnan().any()


assert_tensor_equal = partial(torch.testing.assert_close, rtol=0, atol=0)
is_tensor_equal = partial(torch.allclose, rtol=0, atol=0)


def has_complete_backbone(
    x: Union[AtomTensor, BackboneTensor],
    fill_value: float = 1e-5,
    backbone_indices: List[int] = [0, 1, 2, 3],
) -> bool:
    """
    Checks a protein coordinate tensor (``L x Num Atoms (default 37) x 3``) for
    a complete backbone assignment.

    If ``fill_value`` entries are found in the ``backbone_indices`` (dim 1) then
    the backbone is not considered complete and we return ``False``. If no
    entries with the ``fill_value`` are found, we return ``True``.

    :param x: Atom Tensor to check backbone completeness.
    :type x: graphein.protein.tensor.types.AtomTensor
    :param fill_value: Fill value used in the ``AtomTensor``. Default is ``1e-5``.
    :type fill_value: Float
    :param backbone_indices: List of indices in dimension 1 of the AtomTensor
        to be checked. Defaults to ``[0, 1, 2, 3]`` for ``[N, Ca, C, O]`` in the
        default assignment.
    :type backbone_indices: List[int].
    :return: Boolean indicating whether the backbone is complete.
    :rtype: bool

    .. seealso::

        :func:`graphein.protein.tensor.testing.has_complete_residue`
        :func:`graphein.protein.tensor.testing.is_complete_structure`


    """
    indices = torch.tensor(backbone_indices)
    return torch.sum(x[:, indices] == fill_value) == 0


def has_complete_residue(
    x: ResidueTensor, residue_type: str, fill_value: float = 1e-5
) -> bool:
    """Checks whether a residue has all of the requisite atoms.

    :param x: Tensor of shape ``37 x 3`` of atom coordinates.
    :type x: torch.Tensor
    :param residue_type: 1 or 3-letter code of residue type.
    :type residue_type: str
    :param fill_value: Fill value used to denote the absence of an atom,
        defaults to ``1e-5``.
    :type fill_value: float, optional
    :return: Boolean indicating whether the residue has all of the requisite
        atoms.
    :rtype: bool

    .. seealso::

        :func:`graphein.protein.tensor.testing.has_complete_residue`
        :func:`graphein.protein.tensor.testing.is_complete_structure`
    """
    if len(residue_type) == 1:
        residue_type = STANDARD_AMINO_ACID_MAPPING_1_TO_3[residue_type]
    true_residue_indices = get_atom_indices()[residue_type]

    def _get_index(y: torch.Tensor) -> Tuple[int, ...]:
        indices = (
            torch.nonzero(torch.sum(y != fill_value, dim=1))
            .squeeze(1)
            .tolist()
        )
        # remove oxt if present
        oxt_index = ATOM_NUMBERING_MODIFIED["OXT"]
        indices = tuple(i for i in indices if i != oxt_index)
        return indices

    present_indices = _get_index(x)
    return present_indices == true_residue_indices


def is_complete_structure(
    x: AtomTensor, residues: Union[List[str], str], fill_value: float = 1e-5
) -> bool:
    """Checks whether a protein structure has all of the requisite atoms.

    :param x: AtomTensor to check for completeness.
    :type x: graphein.protein.tensor.types.AtomTensor
    :param residues: List of 1 or 3-letter codes of residue types.
    :type residues: Union[List[str], str]
    :param fill_value: Fill value used to denote the absence of an atom,
        defaults to ``1e-5``.
    :type fill_value: float, optional
    :return: Boolean indicating whether the structure has all of the requisite
        atoms.
    :rtype: bool

    .. seealso::

        :func:`graphein.protein.tensor.sequence.infer_residue_types`
    """
    length = x.shape[0]
    assert (
        len(residues) == length
    ), "Length of sequence must match coordinate tensor in dimension 0."

    if isinstance(residues, str):
        residues = list(residues)

    return all(
        has_complete_residue(x[i], residues[i], fill_value=fill_value)
        for i in range(length)
    )


def random_atom_tensor(length: int = 64) -> AtomTensor:
    """Returns a random tensor of shape ``length x 37 x 3``.


    .. seealso::

        :func:`random_coord_tensor`
        :func:`graphein.protein.tensor.data.get_random_protein`
        :func:`graphein.protein.tensor.data.get_random_batch`

    :param length: Length of random protein, defaults to ``64``
    :type length: int, optional
    :return: Random tensor of shape ``length x 37 x 3``
    :rtype: AtomTensor
    """
    return torch.rand((length, 37, 3))


def random_coord_tensor(length: int = 64) -> CoordTensor:
    """Returns a random tensor of shape ``length x 3``.

    .. seealso::
        :func:`random_atom_tensor`
        :func:`graphein.protein.tensor.data.get_random_protein`
        :func:`graphein.protein.tensor.data.get_random_batch`

    :param length: Length of random coordinates, defaults to ``64``
    :type length: int, optional
    :return: Random tensor of shape ``length x 3``
    :rtype: CoordTensor
    """
    return torch.rand((length, 3))


def fill_missing_o_coords(x: AtomTensor):
    # TODO
    raise NotImplementedError


def fill_missing_backbone(
    x: Union[AtomTensor, CoordTensor], fill_value: float = 1e-5
) -> Union[AtomTensor, BackboneTensor]:
    if has_complete_backbone(x, fill_value):
        return x
    # TODO
    raise NotImplementedError
